import tempfile
import os
import shutil
import subprocess
from .tikz import TikzBackend
from .backend import Backend, Component, Bounds
from typing import List

class PdfBackend(Backend):
    """
    Generates a TikZ picture and compiles it with pdflatex to a .pdf
    Requires pdflatex in the PATH.

    Currently only supports a single figure being created at a time. That is, multi-threaded applications
    need to create one PdfBackend object per figure. Further, if an intermediate directory is set, that directory
    cannot be the same for two figures that are created at the same time, as they might overwrite each other's files.
    """

    def __init__(self, intermediate_dir = None, preamble_lines=[ "\\usepackage[utf8]{inputenc}",
            "\\usepackage[T1]{fontenc}", "\\usepackage{libertine}" ]):
        self._custom_preamble = "\n".join(preamble_lines)
        self._tikz_gen = TikzBackend(False)

        if intermediate_dir is not None:
            self._temp_folder = None
            self._intermediate_dir = intermediate_dir
            os.makedirs(intermediate_dir, exist_ok=True)
        else:
            self._temp_folder = tempfile.TemporaryDirectory()
            self._intermediate_dir = self._temp_folder.name

    def __del__(self):
        if self._temp_folder is not None:
            self._temp_folder.cleanup()
            self._temp_folder = None

    @property
    def preamble(self) -> str:
        return '\n'.join([
            "\\documentclass[varwidth=500cm, border=0pt]{standalone}",
        ])

    def assemble_grid(self, components: List[Component], output_dir: str):
        return self._tikz_gen.assemble_grid(components, self._intermediate_dir)

    def combine_grids(self, data, idx: int, bounds: Bounds):
        return self._tikz_gen.combine_grids(data, idx, bounds)

    def combine_rows(self, data, bounds: Bounds):
        return self._tikz_gen.combine_rows(data, bounds)

    def write_to_file(self, data, filename):
        _, ext = os.path.splitext(filename)
        assert ext.lower() == ".pdf", "Filename should have .pdf extension!"

        tikz_filename = os.path.join(self._intermediate_dir, "figure.tikz")
        self._tikz_gen.write_to_file(data, tikz_filename)

        tex_code = "\n".join([
            self.preamble,
            self._custom_preamble,
            self._tikz_gen.preamble,
            self._tikz_gen.header,
            "\\begin{document}",
            "\\input{figure.tikz}",
            "\\end{document}",
        ])

        tex_filename = os.path.join(self._intermediate_dir, "figure.tex")
        pdf_filename = os.path.join(self._intermediate_dir, "figure.pdf")
        with open(tex_filename, "w") as f:
            f.write(tex_code)

        try:
            subprocess.check_call(['pdflatex', '-interaction=nonstopmode', "figure.tex"],
                cwd=self._intermediate_dir, stdout=subprocess.DEVNULL)
        except subprocess.CalledProcessError:
            print("ERROR: pdflatex reported an error. The .pdf file was copied anyway but likely contains errors.")

        shutil.copy(pdf_filename, filename)


from typing import List, Tuple
from dataclasses import dataclass
import os
from .figuregen import *
from . import calculate as calc
from .element_data import *

@dataclass
class Bounds:
    top: float
    left: float
    width: float
    height: float

@dataclass
class Component:
    bounds: Bounds
    figure_idx: int
    grid_idx: int
    row_idx: int
    col_idx: int

@dataclass
class ImageComponent(Component):
    data: ElementData
    has_frame: bool
    frame_linewidth: float
    frame_color: Tuple[float, float, float]

@dataclass
class TextComponent(Component):
    content: str
    rotation: float
    fontsize: float
    color: Tuple[float, float, float]
    background_color: Tuple[float, float, float]
    type: str
    horizontal_alignment: str = "center"
    padding: calc.Size = calc.Size(0, 0)
    vertical_alignment: str = "center"

@dataclass
class RectangleComponent(Component):
    color: Tuple[float, float, float]
    linewidth: float
    dashed: bool

@dataclass
class LineComponent(Component):
    from_x: float
    from_y: float
    to_x: float
    to_y: float
    linewidth: float
    color: Tuple[float, float, float]

class Backend:
    def generate(self, grids: List[List[Grid]], width_mm: float, filename: str):
        output_dir = os.path.dirname(filename)

        gen_rows = []
        top = 0
        row_idx = 0
        for row in grids:
            sizes = self.compute_aligned_sizes(row, width_mm)

            # generate all grids
            gen_grids = []
            left = 0
            for grid_idx in range(len(row)):
                bounds = Bounds(top, left, sizes[grid_idx][0].width_mm, sizes[grid_idx][0].height_mm)
                components = self.gen_grid(grids[row_idx][grid_idx], bounds, sizes[grid_idx][1])

                # Set the correct figure and grid indices on all components
                for c in components:
                    c.figure_idx = row_idx
                    c.grid_idx = grid_idx

                gen_grids.append(self.assemble_grid(components, output_dir))
                left += sizes[grid_idx][0].width_mm

            bounds = Bounds(top, 0, width_mm, sizes[0][0].height_mm)
            gen_rows.append(self.combine_grids(gen_grids, row_idx, bounds))
            top += sizes[0][0].height_mm
            row_idx += 1

        # Combine all rows
        result = self.combine_rows(gen_rows, Bounds(0, 0, width_mm, top))
        self.write_to_file(result, filename)

    def compute_aligned_sizes(self, grids: List[Grid], width_mm: float) -> List[Tuple[calc.Size, calc.Size]]:
        """
        Computes the sizes of all grids and contained images so that their heights match and they fill the given
        width exactly.

        Returns:
            a list where each element is a tuple of (grid size, image size)
        """
        num_modules = len(grids)
        assert(num_modules != 0)

        if num_modules == 1:
            elem_size = calc.element_size_from_width(grids[0], width_mm)
            h = calc.total_height(grids[0], elem_size)
            return [ (calc.Size(width_mm, h), elem_size) ]

        sum_inverse_aspect_ratios = 0
        inverse_aspect_ratios = []
        for g in grids:
            a = g.rows / g.cols * g.aspect_ratio
            sum_inverse_aspect_ratios += 1/a
            inverse_aspect_ratios.append(1/a)

        sum_fixed_deltas = 0
        i = 0
        for m in grids:
            w_fix = calc.min_width(m)
            h_fix = calc.min_height(m)
            sum_fixed_deltas += w_fix - h_fix * inverse_aspect_ratios[i]
            i += 1

        height = (width_mm - sum_fixed_deltas) / sum_inverse_aspect_ratios

        sizes = []
        for m in grids:
            elem_size = calc.element_size_from_height(m, height)
            w = calc.total_width(m, elem_size)
            sizes.append((calc.Size(w, height), elem_size))
        return sizes

    def gen_lines(self, element: ElementView, img_pos_top, img_pos_left, img_size: calc.Size) -> List[Component]:
        try:
            lines = element.elem['lines']
        except:
            return []

        if lines == []:
            return []

        if isinstance(element.image, RasterImage):
            # Coordinates are in pixels
            w_scale = img_size.width_mm / element.image.width_px
            h_scale = img_size.height_mm / element.image.height_px
        else:
            # Coordinates are between 0 and 1.
            w_scale = img_size.width_mm
            h_scale = img_size.height_mm

        result = []
        for l in lines:
            start_h = l['from'][0] * h_scale + img_pos_top
            start_w = l['from'][1] * w_scale + img_pos_left
            end_h = l['to'][0] * h_scale + img_pos_top
            end_w = l['to'][1] * w_scale + img_pos_left
            rgb = (l['color'][0], l['color'][1], l['color'][2])
            bounds = Bounds(img_pos_top, img_pos_left, img_size.width_mm, img_size.height_mm)
            result.append(LineComponent(bounds, -1, -1, -1, -1, start_w, start_h, end_w, end_h, l['linewidth'], rgb))

        return result

    def _gen_label(self, img_pos_top, img_pos_left, img_width, img_height, cfg, label_pos) -> TextComponent:
        try:
            cfg = cfg[label_pos]
        except KeyError:
            return None

        alignment = label_pos.split('_')[-1]
        is_top = (label_pos.split('_')[0] == 'top')

        rect_width, rect_height = cfg['width_mm'], cfg['height_mm']

        # determine the correct offsets depending on wether it is in the corner or center
        if alignment == 'center':
            offset_w, offset_h = 0, cfg['offset_mm']
        else:
            offset_w = cfg['offset_mm'][0]
            offset_h = cfg['offset_mm'][1]

        # determine pos_top of rectangle
        if is_top:
            pos_top = img_pos_top + offset_h
        else:
            pos_top = img_pos_top + img_height - rect_height - offset_h

        # determine pos_left of rectangle based on alignment
        if alignment == 'center':
            pos_left = img_pos_left + (img_width * 1/2.) - (rect_width * 1/2.)
        elif alignment == 'left':
            pos_left = img_pos_left + offset_w
        else: # right
            pos_left = img_pos_left + img_width - rect_width - offset_w

        bounds = Bounds(pos_top, pos_left, rect_width, rect_height)

        padding = calc.Size(cfg['padding_mm'], cfg['padding_mm'])

        c = TextComponent(bounds, -1, -1, -1, -1, cfg["text"], 0, cfg['fontsize'], cfg['text_color'],
            cfg['background_color'], "label-" + label_pos, alignment, padding, "top" if is_top else "bottom")

        return c

    def gen_labels(self, element: ElementView, img_pos_top, img_pos_left, img_size: calc.Size) -> List[Component]:
        try:
            cfg = element.elem['label']
        except:
            return []

        labels = []
        for label_pos in ['top_center', 'top_left', 'top_right', 'bottom_center', 'bottom_left', 'bottom_right']:
            l = self._gen_label(img_pos_top, img_pos_left, img_size.width_mm, img_size.height_mm, cfg, label_pos)
            if l is None:
                continue
            labels.append(l)
        return labels

    def gen_markers(self, element: ElementView, img_pos_top, img_pos_left, img_size: calc.Size) -> List[Component]:
        try:
            markers = element.elem['crop_marker']
        except:
            return []

        if isinstance(element.image, RasterImage):
            # Coordinates are in pixels
            w_scale = img_size.width_mm / element.image.width_px
            h_scale = img_size.height_mm / element.image.height_px
        else:
            # Coordinates are between 0 and 1.
            w_scale = img_size.width_mm
            h_scale = img_size.height_mm

        result = []
        for m in markers:
            if m['linewidth'] > 0.0:
                pos_top = img_pos_top + (m['pos'][1] * h_scale)
                pos_left = img_pos_left + (m['pos'][0] * w_scale)
                w = m['size'][0] * w_scale
                h = m['size'][1] * h_scale
                bounds = Bounds(pos_top, pos_left, w, h)
                result.append(RectangleComponent(bounds, -1, -1, -1, -1, m['color'], m['linewidth'], m['dashed']))
        return result

    def gen_images(self, grid: Grid, grid_bounds: Bounds, img_size: calc.Size) -> List[Component]:
        """ Generates a list of figure components for all images and their lables, captions, frames, and markers """
        images = []
        for row_idx in range(grid.rows):
            for col_idx in range(grid.cols):
                element = grid[row_idx, col_idx]
                assert element.image is not None
                assert isinstance(element.image, ElementData)

                # Position of the main image
                pos_top, pos_left = calc.image_pos(grid, img_size, col_idx, row_idx)
                pos_top += grid_bounds.top
                pos_left += grid_bounds.left
                bounds = Bounds(pos_top, pos_left, img_size.width_mm, img_size.height_mm)

                # If there is a frame, get is properties
                has_frame = "frame" in element.elem and element.elem["frame"] is not None
                linewidth = 0
                color = [0,0,0]
                if has_frame:
                    linewidth = element.elem["frame"]["line_width"]
                    color = element.elem["frame"]["color"]

                images.append(ImageComponent(bounds, -1, -1, row_idx, col_idx, element.image,
                    has_frame, linewidth, color))

                markers = self.gen_lines(element, pos_top, pos_left, img_size)
                for m in markers:
                    m.row_idx = row_idx
                    m.col_idx = col_idx
                images.extend(markers)

                markers = self.gen_labels(element, pos_top, pos_left, img_size)
                for m in markers:
                    m.row_idx = row_idx
                    m.col_idx = col_idx
                images.extend(markers)

                markers = self.gen_markers(element, pos_top, pos_left, img_size)
                for m in markers:
                    m.row_idx = row_idx
                    m.col_idx = col_idx
                images.extend(markers)
        return images

    def gen_south_captions(self, grid: Grid, grid_bounds: Bounds, img_size: calc.Size) -> List[Component]:
        layout = grid.layout.layout['element_config']['captions']['south']
        if layout['height'] == 0:
            return []

        captions = []
        for row_idx in range(grid.rows):
            for col_idx in range(grid.cols):
                if 'captions' not in grid[row_idx, col_idx].elem:
                    continue

                txt_content = grid[row_idx, col_idx].elem['captions']['south']

                if txt_content == '':
                    continue

                (top, left) = calc.south_caption_pos(grid, img_size, col_idx, row_idx)
                bounds = Bounds(top + grid_bounds.top, left + grid_bounds.left,
                    img_size.width_mm, layout['height'])

                captions.append(TextComponent(bounds, -1, -1, row_idx, col_idx, txt_content, layout['rotation'],
                    layout['fontsize'], layout['text_color'], [255, 255, 255], "caption", vertical_alignment="top"))

        return captions

    def gen_titles(self, grid: Grid, grid_bounds: Bounds, img_size: calc.Size) -> List[Component]:
        titles = []
        for direction in ['north', 'east', 'south', 'west']:
            if direction not in grid.data['titles']:
                continue

            content = grid.data['titles'][direction]
            (top, left, width, height) = calc.titles_pos_and_size(grid, img_size, direction)
            bounds = Bounds(top + grid_bounds.top, left + grid_bounds.left, width, height)

            if width == 0 or height == 0 or content == "":
                continue

            t = grid.layout.layout['titles'][direction]
            titles.append(TextComponent(bounds, -1, -1, -1, -1, content, t['rotation'], t['fontsize'],
                t['text_color'], t['background_color'], "title-" + direction))
        return titles

    def _compute_bg_colors(self, bg_color_properties, num):
        if bg_color_properties is None: # no background color
            return [None for i in range(num)]
        elif not isinstance(bg_color_properties[0], list): # constant color for all
            return [bg_color_properties for i in range(num)]
        else: # individual background colors
            return bg_color_properties

    def _gen_row_col_titles(self, direction: str, layout, num: int, pos_fn, contents: List[str], is_row):
        titles = []
        if calc.size_of(layout, direction)[0] != 0.0:
            bg_colors = self._compute_bg_colors(layout[direction]['background_colors'], num)
            t = layout[direction]

            for i in range(num):
                bounds = pos_fn(i)
                if bounds.width == 0 or bounds.height == 0:
                    continue

                txt = contents[i]
                if txt == "":
                    continue

                if is_row:
                    titles.append(TextComponent(bounds, -1, -1, i, -1, txt, t['rotation'], t['fontsize'],
                        t['text_color'], bg_colors[i], "rowtitle-" + direction))
                else:
                    titles.append(TextComponent(bounds, -1, -1, -1, i, txt, t['rotation'], t['fontsize'],
                        t['text_color'], bg_colors[i], "coltitle-" + direction))
        return titles

    def gen_row_titles(self, grid: Grid, grid_bounds: Bounds, img_size: calc.Size) -> List[Component]:
        titles = []
        for direction in ['east', 'west']:
            def pos_fn(idx):
                (top, left, width, height) = calc.row_titles_pos(grid, img_size, idx + 1, direction)
                return Bounds(top + grid_bounds.top, left + grid_bounds.left, width, height)

            if direction not in grid.data['row_titles']:
                continue

            contents = grid.data['row_titles'][direction]['content']
            t = self._gen_row_col_titles(direction, grid.layout.layout['row_titles'], grid.rows, pos_fn,
                contents, True)
            titles.extend(t)

        return titles

    def gen_column_titles(self, grid: Grid, grid_bounds: Bounds, img_size: calc.Size) -> List[Component]:
        titles = []
        for direction in ['north', 'south']:
            def pos_fn(idx):
                (top, left, width, height) = calc.column_titles_pos(grid, img_size, idx + 1, direction)
                return Bounds(top + grid_bounds.top, left + grid_bounds.left, width, height)

            if direction not in grid.data['column_titles']:
                continue

            contents = grid.data['column_titles'][direction]['content']
            t = self._gen_row_col_titles(direction, grid.layout.layout['column_titles'], grid.cols, pos_fn,
                contents, False)
            titles.extend(t)

        return titles

    def gen_grid(self, grid: Grid, bounds: Bounds, img_size: calc.Size) -> List[Component]:
        result = []
        result.extend(self.gen_images(grid, bounds, img_size))
        result.extend(self.gen_south_captions(grid, bounds, img_size))
        result.extend(self.gen_titles(grid, bounds, img_size))
        result.extend(self.gen_row_titles(grid, bounds, img_size))
        result.extend(self.gen_column_titles(grid, bounds, img_size))
        return result

    def assemble_grid(self, components: List[Component], output_dir: str):
        raise NotImplementedError()

    def combine_grids(self, data, idx: int, bounds: Bounds):
        raise NotImplementedError()

    def combine_rows(self, data, bounds: Bounds):
        raise NotImplementedError()

    def write_to_file(self, data, filename):
        raise NotImplementedError()



from typing import Tuple, List
import numpy as np

from . import image
from .. import figuregen

class CropComparison:
    """ Matrix of cropped and zoomed images next to a reference image.

    Derived classes can change some behaviour, like the choice of error metric, by overriding the
    corresponding methods.

    Additional content can be added (or removed) by the user simply by accessing the individual
    grids in the generated list of grids.
    """
    def __init__(self, reference_image, method_images, crops: List[image.Cropbox],
                 scene_name = None, method_names = None, use_latex = False):
        """ Shows a reference image next to a grid of crops from different methods.

        Args:
            reference_image: a reference image (or any other image to put full-size in the lefthand grid)
            method_images: list of images, each corresponds to a new column in the crop grid
            crops: list of crops to take from each method, each creates a new row and a marker on the reference
            scene_name: [optional] string, name of the scene to put underneath the reference image
            method_names: [optional] list of string, names for the reference and each method, to put above the crops
            use_latex: set to true to pretty-print captions with LaTeX commands (requires TikZ backend)

        Returns:
            A list of two grids:
            The first is a single image (reference), the second a series of crops, one or more for each method.
        """
        self._reference_image = reference_image
        self._method_images = method_images
        self.use_latex = use_latex

        self._errors = [
            self.compute_error(reference_image, m)
            for m in method_images
        ]
        self._crop_errors = [
            [
                self.compute_error(crop.crop(reference_image), crop.crop(m))
                for m in method_images
            ]
            for crop in crops
        ]

        # Create the grid for the reference image
        self._ref_grid = figuregen.Grid(1, 1)
        self._ref_grid[0, 0].image = self.tonemap(reference_image)
        for crop in crops:
            self._ref_grid[0, 0].set_marker(crop.marker_pos, crop.marker_size, color=[255,255,255])

        if scene_name is not None:
            self._ref_grid.set_col_titles("bottom", [scene_name])

        # Create the grid with the crops
        self._crop_grid = figuregen.Grid(num_cols=len(method_images) + 1, num_rows=len(crops))
        for row in range(len(crops)):
            self._crop_grid[row, 0].image = self.tonemap(crops[row].crop(reference_image))
            for col in range(len(method_images)):
                self._crop_grid[row, col + 1].image = self.tonemap(crops[row].crop(method_images[col]))

        # Put error values underneath the columns
        error_strings = [ f"{self.error_metric_name}" ]
        error_strings.extend([ self.error_string(i, self.errors) for i in range(len(self.errors)) ])
        self._crop_grid.set_col_titles("bottom", error_strings)
        self._crop_grid.layout.set_padding(column=1, row=1)
        self._crop_grid.layout.set_col_titles("bottom", fontsize=8, field_size_mm=2.8, offset_mm=0.5)

        # If given, show method names on top
        if method_names is not None:
            self._crop_grid.set_col_titles("top", method_names)
            self._crop_grid.layout.set_col_titles("top", fontsize=8, field_size_mm=2.8, offset_mm=0.25)

        self._ref_grid.copy_layout(self._crop_grid)
        self._ref_grid.layout.set_padding(right=1)

    def tonemap(self, img):
        return figuregen.JPEG(image.lin_to_srgb(img), quality=80)

    @property
    def error_metric_name(self) -> str:
        return "relMSE"

    def compute_error(self, reference_image, method_image) -> Tuple[str, List[float]]:
        return image.relative_mse(method_image, reference_image, 0.01)

    def error_string(self, index: int, errors: List[float]):
        """ Generates the human-readable error string for the i-th element in a list of error values.

        Args:
            index: index in the list of errors
            errors: list of error values, one per method, in order
        """
        if self.use_latex and index == np.argmin(errors):
            return f"$\\mathbf{{{errors[index]:.2f} ({errors[index]/errors[0]:.2f}\\times)}}$"
        elif self.use_latex:
            return f"${errors[index]:.2f} ({errors[index]/errors[0]:.2f}\\times)$"
        else:
            return f"{errors[index]:.2f} ({errors[index]/errors[0]:.2f}x)"

    @property
    def crop_errors(self) -> List[List[float]]:
        """ Error values within the cropped region of each method.
        First dimension is the crop, second the method.
        """
        return self._crop_errors

    @property
    def errors(self) -> List[float]:
        return self._errors

    @property
    def figure_row(self) -> List[figuregen.Grid]:
        return [ self._ref_grid, self._crop_grid ]

class FullSizeWithCrops:
    """ Side-by-side comparison of full-size images. Below each image is a row of crops.

    Derived classes can change some behaviour, like the choice of error metric, by overriding the
    corresponding methods.

    Additional content can be added (or removed) by the user simply by accessing the individual
    grids in the generated list of grids.
    """
    def __init__(self, reference_image, method_images, crops: List[image.Cropbox],
                 crops_below = True, method_names = None, use_latex = False):
        """ Shows a reference image next to a grid of crops from different methods.

        Args:
            reference_image: a reference image (or any other image to put full-size in the lefthand grid)
            method_images: list of images, each corresponds to a new column in the crop grid
            crops: list of crops to take from each method, each creates a new row and a marker on the reference
            crops_below: [optional] if False, the crops will be a column to the right of each image
            method_names: [optional] list of string, names for the reference and each method, to put above the crops
            use_latex: set to true to pretty-print captions with LaTeX commands (requires TikZ backend)

        Returns:
            A list of two grids:
            The first is a single image (reference), the second a series of crops, one or more for each method.
        """
        self._reference_image = reference_image
        self._method_images = method_images
        self.use_latex = use_latex
        self._crops_below = crops_below

        self._errors = [
            self.compute_error(reference_image, m)
            for m in method_images
        ]
        self._crop_errors = [
            [
                self.compute_error(crop.crop(reference_image), crop.crop(m))
                for m in method_images
            ]
            for crop in crops
        ]

        # Put in one list to make our life easier in the following
        images = [reference_image]
        images.extend(method_images)

        # Create the grid for the reference image
        self._ref_grid = [ figuregen.Grid(1, 1) for _ in range(len(images)) ]
        for i in range(len(images)):
            self._ref_grid[i][0, 0].image = self.tonemap(images[i])
            for crop in crops:
                self._ref_grid[i][0, 0].set_marker(crop.marker_pos, crop.marker_size, color=[255,255,255])

        # Create the grid with the crops
        if self._crops_below:
            self._crop_grid = [
                figuregen.Grid(num_cols=len(crops), num_rows=1)
                for _ in range(len(images))
            ]
            for i in range(len(images)):
                for col in range(len(crops)):
                    self._crop_grid[i][0, col].image = self.tonemap(crops[col].crop(images[i]))
        else:
            self._crop_grid = [
                figuregen.Grid(num_cols=1, num_rows=len(crops))
                for _ in range(len(images))
            ]
            for i in range(len(images)):
                for row in range(len(crops)):
                    self._crop_grid[i][row, 0].image = self.tonemap(crops[row].crop(images[i]))

        # Add padding to the right of all but the last image
        for i in range(len(images) - 1):
            self._ref_grid[i].layout.set_padding(right=1)
            self._crop_grid[i].layout.set_padding(right=1)
            if self._crops_below:
                self._ref_grid[i].layout.set_padding(bottom=1)

        if self._crops_below:
            self._ref_grid[-1].layout.set_padding(bottom=1)
        else:
            self._ref_grid[-1].layout.set_padding(right=1)

        # Put error values underneath the columns
        if self._crops_below:
            for i in range(len(images)):
                if i > 0:
                    err = self.error_string(i - 1, self.errors)
                else:
                    err = self.error_metric_name
                self._crop_grid[i].set_title("bottom", method_names[i] + "\\\\" + err)
                self._crop_grid[i].layout.set_title("bottom", 6, 1, 8)
        else:
            pass # TODO

        # TODO this requires titles spanning multiple grids (the image and its crops)!
        # error_strings = [ f"{self.error_metric_name}" ]
        # error_strings.extend([ self.error_string(i, self.errors) for i in range(len(self.errors)) ])
        # self._crop_grid.set_col_titles("bottom", error_strings)
        # self._crop_grid.layout.set_padding(column=1, row=1)
        # self._crop_grid.layout.set_col_titles("bottom", fontsize=8, field_size_mm=2.8, offset_mm=0.5)

        # If given, show method names on top
        # TODO combine with error values, and always show both or neither
        # if method_names is not None:
        #     self._crop_grid.set_col_titles("top", method_names)
        #     self._crop_grid.layout.set_col_titles("top", fontsize=8, field_size_mm=2.8, offset_mm=0.25)

        # self._ref_grid.copy_layout(self._crop_grid)
        # self._ref_grid.layout.set_padding(right=1)
        # TODO set appropriate paddings for alignment etc

    def tonemap(self, img):
        return figuregen.JPEG(image.lin_to_srgb(img), quality=80)

    @property
    def error_metric_name(self) -> str:
        return "relMSE"

    def compute_error(self, reference_image, method_image) -> Tuple[str, List[float]]:
        return image.relative_mse(method_image, reference_image, 0.01)

    def error_string(self, index: int, errors: List[float]):
        """ Generates the human-readable error string for the i-th element in a list of error values.

        Args:
            index: index in the list of errors
            errors: list of error values, one per method, in order
        """
        if self.use_latex and index == np.argmin(errors):
            return f"$\\mathbf{{{errors[index]:.2f} ({errors[index]/errors[0]:.2f}\\times)}}$"
        elif self.use_latex:
            return f"${errors[index]:.2f} ({errors[index]/errors[0]:.2f}\\times)$"
        else:
            return f"{errors[index]:.2f} ({errors[index]/errors[0]:.2f}x)"

    @property
    def crop_errors(self) -> List[List[float]]:
        """ Error values within the cropped region of each method.
        First dimension is the crop, second the method.
        """
        return self._crop_errors

    @property
    def errors(self) -> List[float]:
        return self._errors

    @property
    def figure(self) -> List[List[figuregen.Grid]]:
        if self._crops_below:
            return [ self._ref_grid, self._crop_grid ]
        else:
            grids = []
            for i in range(len(self._method_images) + 1):
                grids.append(self._ref_grid[i])
                grids.append(self._crop_grid[i])
            return [ grids ]


from typing import Union
from .backend import *
from concurrent.futures import ThreadPoolExecutor, Future

class HtmlBackend(Backend):
    def __init__(self, inline=False, custom_head: str = "", id_prefix=""):
        """
        Creates a new HTML backend that emits a static webpage with embedded base64 images.

        Args:
            inline: boolean, if true does not generate <html>, <head>, <body> tags
            custom_head: additonal lines to add to the <head>, ignored if inline = False
            id_prefix: additional text in front of the figure's elements, useful if multiple figures are in on .html
        """
        self._inline = inline
        self._custom_head = custom_head
        self._prefix = id_prefix
        self._thread_pool = ThreadPoolExecutor()

    @property
    def style(self) -> str:
        """ The required CSS style for the figure to display correctly. """
        return "\n".join([
            '<style type="text/css">',
            '.figure { position: relative; }',
            '.title-container { position: absolute; margin: 0; }',
            '.title-content { margin: 0; }',
            '.element { position: absolute; }',
            '.topalign { display: flex; justify-content: flex-start; flex-direction: column; } ',
            '.botalign { display: flex; justify-content: flex-end; flex-direction: column;  } ',
            '.centeralign { display: flex; justify-content: center; flex-direction: column;  } ',
            'body { margin: 0; }'
            '</style>',
        ])

    def _html_color(self, rgb) -> str:
        return f"rgb({rgb[0]},{rgb[1]},{rgb[2]})"

    def _make_image(self, c: ImageComponent, dims, pos, elem_id):
        # Generate the image data
        elem_idx = self._prefix + "img-" + elem_id
        imgtag = c.data.make_html(c.bounds.width, c.bounds.height)

        html_code = f"<div class='element' id='{elem_idx}' style='"
        html_code += dims + pos

        # Check if there is a frame and emit the correct command
        frame = ""
        if c.has_frame:
            frame += f"<div style='position: absolute; top: 0; left: 0; " + dims
            frame += f"border: {c.frame_linewidth}pt solid "
            frame += f"{self._html_color(c.frame_color)}; box-sizing: border-box;'></div>"

        html_code += "' >" + imgtag + frame + "</div>"
        return html_code

    def assemble_grid(self, components: List[Component], output_dir: str):
        html_lines = []
        for c in components:
            elem_id = f"fig{c.figure_idx}-grid{c.grid_idx}"
            if c.row_idx >= 0:
                elem_id += f"-row{c.row_idx}"
            if c.col_idx >= 0:
                elem_id += f"-col{c.col_idx}"

            # Position arguments are the same for all components
            if c.bounds is not None:
                pos = f"top: {c.bounds.top}mm; left: {c.bounds.left}mm; "
                dims = f"width: {c.bounds.width}mm; height: {c.bounds.height}mm;"

            if isinstance(c, ImageComponent):
                html_lines.append(self._thread_pool.submit(self._make_image, c, dims, pos, elem_id))

            if isinstance(c, TextComponent):
                elem_idx = self._prefix + c.type + "-" + elem_id

                color = "color: " + self._html_color(c.color) + "; "
                fontsize = "font-size: " + f'{c.fontsize}pt' + "; "
                horz_align = "text-align: " + c.horizontal_alignment + "; "

                rotation = ""
                if c.rotation == -90:
                    rotation = f"transform: rotate(90deg) translateX({-c.bounds.width}mm);"
                    rotation += "transform-origin: bottom left;"
                elif c.rotation == 90:
                    rotation = f"transform: rotate(-90deg) translateX({-c.bounds.height}mm);"
                    rotation += "transform-origin: top left;"

                # Need to flip the dimensions if we rotate
                if rotation != "":
                    dims = f"width: {c.bounds.height}mm; height: {c.bounds.width}mm;"

                if c.background_color is not None:
                    bgn = "background-color: " + self._html_color(c.background_color) + "; "
                else:
                    bgn = ""
                pad = f"padding-top: {c.padding.height_mm}mm; padding-bottom: {c.padding.height_mm}mm; "
                pad += f"padding-left: {c.padding.width_mm}mm; padding-right: {c.padding.width_mm}mm; "
                pad += "box-sizing: border-box; "

                aligncls = "centeralign"
                if c.vertical_alignment == "top":
                    aligncls = "topalign"
                elif c.vertical_alignment == "bottom":
                    aligncls = "botalign"

                html_lines.append("\n".join([
                    f"<div class='title-container {aligncls}' id='{elem_idx}' style='" + dims + pos
                        + pad + bgn + rotation + "' >",
                    "<p class='title-content' style='" + color + fontsize + horz_align + "'>",
                    c.content.replace('\\\\', '<br/>'),
                    "</p>",
                    "</div>"
                ]))

            if isinstance(c, RectangleComponent):
                html_code = "<div style='position: absolute; " + dims + pos
                html_code += f"border: {c.linewidth}pt {self._html_color(c.color)}"
                html_code += 'dashed' if c.dashed else 'solid' + "; "
                html_code += "'></div>"
                html_lines.append(html_code)

            if isinstance(c, LineComponent):
                html_code = f'<div class="svg-container" style="position: absolute; {dims + pos}">'
                html_code += f'<svg style="{dims}">'
                html_code += f'<line x1="{c.from_x - c.bounds.left}mm" y1="{c.from_y - c.bounds.top}mm" '
                html_code += f'x2="{c.to_x - c.bounds.left}mm" y2="{c.to_y - c.bounds.top}mm" '
                html_code += f'style="stroke:{self._html_color(c.color)}; stroke-width:{c.linewidth}pt" />'
                html_code += '</svg></div>'
                html_lines.append(html_code)

        return html_lines

    def combine_grids(self, data: List[List[Union[str, Future]]], idx: int, bounds: Bounds) -> List[Union[str, Future]]:
        # Create a container div for each row
        figure_id = self._prefix + "figure-" + str(idx)
        pos = f"top: {bounds.top}mm; left: {bounds.left}mm; "
        dims = f"width: {bounds.width}mm; height: {bounds.height}mm; "

        # Flatten the inner lines and combine
        result = ["<div id='" + figure_id + "' style='" + pos + dims + "'>"]
        for grid in data:
            for line in grid:
                result.append(line)
        result.append("</div>")

        return result

    def combine_rows(self, data: List[List[Union[str, Future]]], bounds: Bounds) -> str:
        # Create a container div to make sure that everything can be moved around on a final page
        pos = f"top: {bounds.top}mm; left: {bounds.left}mm; "
        dims = f"width: {bounds.width}mm; height: {bounds.height}mm; "

        # Synchronize all export tasks
        html_code = "<div class='figure' style='" + pos + dims + "'>\n"
        for row in data:
            for line in row:
                if isinstance(line, Future):
                    html_code += line.result() + "\n"
                else:
                    html_code += line + "\n"
        html_code += "</div>\n"
        return html_code

    def write_to_file(self, data: str, filename: str):
        with open(filename, "w") as f:
            if not self._inline:
                f.writelines([
                    "<!DOCTYPE html>",
                    "<html>",
                    "<head>",
                ])
                f.write(self.style)
                f.write(self._custom_head)
                f.writelines([
                    "</head>"
                    "<body>"
                ])

            f.write(data)

            if not self._inline:
                f.write("</body>")



layouts = {
    "grid": {
        "units": "mm",
        "LaTeX_font_family": "libertine",
        "background_color": [ 255, 255, 255 ],

        "row_space": 0.8,
        "column_space": 0.8,

        "padding": {
            "north": 0.0,
            "west": 0.0,
            "east": 0.0,
            "south": 0.0
        },

        "element_config": {
            "captions": {
                "north": {
                    "height": 0.0,
                    "offset": 0.0,
                    "rotation": 0,
                    "fontsize": 6,
                    "line_space": 1.2,
                    "text_color": [ 0, 0, 0 ]
                },

                "south": {
                    "height": 0.0,
                    "offset": 0.0,
                    "rotation": 0,
                    "fontsize": 6,
                    "line_space": 1.2,
                    "text_color": [ 0, 0, 0 ]
                },

                "east": {
                    "width": 0.0,
                    "offset": 0.0,
                    "rotation": -90,
                    "fontsize": 6,
                    "line_space": 1.2,
                    "text_color": [ 0, 0, 0 ]
                },

                "west": {
                    "width": 0.0,
                    "offset": 0.0,
                    "rotation": 90,
                    "fontsize": 6,
                    "line_space": 1.2,
                    "text_color": [ 0, 0, 0 ]
                }
            }
        },
        "titles": {
            "north": {
                "height": 0.0,
                "offset": 0.0,
                "rotation": 0,
                "fontsize": 8,
                "line_space": 1.2,
                "text_color": [ 0, 0, 0 ],
                "background_color": None
            },
            "south": {
                "height": 0.0,
                "offset": 0.0,
                "rotation": 0,
                "fontsize": 8,
                "line_space": 1.2,
                "text_color": [ 0, 0, 0 ],
                "background_color": None
            },
            "west": {
                "width": 0.0,
                "offset": 0.0,
                "rotation": 90,
                "fontsize": 8,
                "line_space": 1.2,
                "text_color": [ 0, 0, 0 ],
                "background_color": None
            },
            "east": {
                "width": 0.0,
                "offset": 0.0,
                "rotation": -90,
                "fontsize": 8,
                "line_space": 1.2,
                "text_color": [ 0, 0, 0 ],
                "background_color": None
            }
        },
        "row_titles": {
            "west": {
                "width": 0.0,
                "offset": 0.0,
                "rotation": 90,
                "fontsize": 7,
                "line_space": 1.2,
                "text_color": [ 0, 0, 0 ],
                "background_colors": None
            },
            "east": {
                "width": 0.0,
                "offset": 0.0,
                "rotation": -90,
                "fontsize": 7,
                "line_space": 1.2,
                "text_color": [ 0, 0, 0 ],
                "background_colors": None
            }
        },
        "column_titles": {
            "north": {
                "height": 0.0,
                "offset": 0.0,
                "rotation": 0,
                "fontsize": 7,
                "line_space": 1.2,
                "text_color": [ 0, 0, 0 ],
                "background_colors": None
            },
            "south": {
                "height": 0.0,
                "offset": 0.0,
                "rotation": 0,
                "fontsize": 7,
                "line_space": 1.2,
                "text_color": [ 0, 0, 0 ],
                "background_colors": None
            }
        }
    }
}


import cv2
import os
import base64
import simpleimageio
import numpy as np
import tempfile

class Error(Exception):
    def __init__(self, message):
        self.message = message

class ElementData:
    @property
    def aspect_ratio(self):
        """ Aspect ratio (height / width) of the grid element. """
        pass

    def make_raster(self, width, height, base_filename) -> str:
        ''' Writes the element to a raster file format (e.g, .png, .jpg)

        This function should always be implemented. All backends use it as the fallback option.
        The exact file format is implementation specific and might be controlled by the user.

        Args:
            width: Desired width of the image in [mm]
            height: Desired height of the image in [mm]
            base_filename: Backend-generated filename prefix that can be used to generate a unique file
                in the correct location

        Returns:
            The filename of the generated file.
        '''
        raise NotImplementedError()

    def make_pdf(self, width, height, base_filename) -> str:
        ''' Writes the element to a .pdf file.

        Args:
            width: Desired width of the image in [mm]
            height: Desired height of the image in [mm]
            base_filename: Backend-generated filename prefix that can be used to generate a unique file
                in the correct location

        Returns:
            The filename of the generated file.
        '''
        raise NotImplementedError()

    def make_html(self, width, height) -> str:
        ''' Writes the element to a static .html page, images should be embedded as base64.

        The default generates a raster image in a temporary file and encodes it as base64 .png

        Args:
            width: Desired width of the image in [mm]
            height: Desired height of the image in [mm]

        Returns:
            The generated inline html code.
        '''
        temp_folder = tempfile.TemporaryDirectory()
        fname = self.make_raster(width, height, os.path.join(temp_folder.name, "image"))
        with open(fname, "rb") as f:
            b64 = base64.b64encode(f.read())
        temp_folder.cleanup()

        html = "<img src='data:image/png;base64," + b64.decode('utf-8')
        html += f"' style='width: {width}mm; height: {height}mm;' />"
        return html

class Plot(ElementData):
    ''' Base class for all generated images and plots.

    Offers a user-controllable aspect ratio
    '''
    @property
    def aspect_ratio(self):
        return self._aspect

    @aspect_ratio.setter
    def aspect_ratio(self, v):
        self._aspect = v

class Image(ElementData):
    @property
    def width_px(self):
        raise NotImplementedError()

    @property
    def height_px(self):
        raise NotImplementedError()

class PDF(Image):
    ''' Loads and embeds the first page of a pdf file.

    Additional dependencies: PyPDF2 and pdf2image (requires poppler to be installed and in the PATH)
    '''
    def __init__(self, filename, dpi=300, use_jpeg=False):
        self.file = filename
        self.dpi = dpi
        self.ext = '.jpg' if use_jpeg else '.png'
        self.mimetype = 'data:image/jpeg;base64,' if use_jpeg else 'data:image/png;base64,'

    @Image.aspect_ratio.getter
    def aspect_ratio(self):
        from PyPDF2 import PdfFileReader
        box = PdfFileReader(open(self.filename, "rb")).getPage(0).mediaBox
        width_pt = box.upperRight[0]
        height_pt = box.upperRight[1]
        return float(float(height_pt) / float(width_pt))

    def convert(self):
        from pdf2image import convert_from_path
        images = convert_from_path(self.file, dpi=self.dpi, last_page=1)
        return np.array(images[0])

    def make_raster(self, width, height, base_filename) -> str:
        img = self.convert()
        cv2.imwrite(base_filename + self.ext, cv2.cvtColor(img, cv2.COLOR_RGB2BGR))
        return base_filename + self.ext

    def make_pdf(self, width, height, base_filename) -> str:
        return self.file

class RasterImage(Image):
    ''' Abstract base class for all supported raster image types. '''
    def __init__(self, raw_image_or_filename):
        '''
            Either provide raw image data OR a filename.
        '''
        assert raw_image_or_filename is not None

        if isinstance(raw_image_or_filename, str):
            self.file = raw_image_or_filename
            self.raw = cv2.imread(self.file)
            self.width = self.raw.shape[1]
            self.height = self.raw.shape[0]
        else:
            self.file = None
            self.raw = raw_image_or_filename
            self.width = self.raw.shape[1]
            self.height = self.raw.shape[0]

    @Image.width_px.getter
    def width_px(self):
        return self.width

    @Image.height_px.getter
    def height_px(self):
        return self.height

    @Image.aspect_ratio.getter
    def aspect_ratio(self):
        return float(self.height / float(self.width))

    def convert(self, out_filename):
        if self.file is None:
            clipped = simpleimageio.to_byte_image(self.raw)
            cv2.imwrite(out_filename, cv2.cvtColor(clipped.astype('uint8'), cv2.COLOR_RGB2BGR))
        else:
            cv2.imwrite(out_filename, self.raw)

class PNG(RasterImage):
    ''' A raster image that will be converted to .png '''
    def __init__(self, raw_image_or_filename):
        self.ext = ".png"
        RasterImage.__init__(self, raw_image_or_filename)

    def make_raster(self, width, height, base_filename) -> str:
        filename = base_filename + self.ext
        self.convert(filename)
        return filename

class JPEG(RasterImage):
    ''' A raster image that will be converted to .jpg '''
    def __init__(self, raw_image_or_filename, quality=85):
        self.ext = ".jpg"
        self.quality = quality
        RasterImage.__init__(self, raw_image_or_filename)

    def make_raster(self, width, height, base_filename) -> str:
        filename = base_filename + self.ext
        clipped = simpleimageio.to_byte_image(self.raw)
        cv2.imwrite(filename, cv2.cvtColor(clipped.astype('uint8'), cv2.COLOR_RGB2BGR),
            [int(cv2.IMWRITE_JPEG_QUALITY), self.quality])
        return filename

class HTML(Image):
    ''' Embeds a .html.

    Currently ownly useful to the HTML backend, as it does not render the webpage.
    '''
    def __init__(self, filename, aspect_ratio):
        self.file = filename
        self.a_ratio = float(aspect_ratio)

    @Image.aspect_ratio.getter
    def aspect_ratio(self):
        return self.a_ratio



import numpy as np

from simpleimageio import lin_to_srgb, luminance, exposure, average_color_channels, zoom

def crop(img, *crop_args):
    '''
    crop_args: list or tuple of 4 elements: left, top, width, height.
    '''
    left, top, width, height = crop_args

    assert top >= 0 and left >= 0, "crop is outside the image"
    assert left + width <= img.shape[1], "crop is outside the image"
    assert top + height <= img.shape[0], "crop is outside the image"

    return img[top:top+height,left:left+width,:]

class Cropbox:
    def __init__(self, top, left, height, width, scale=1):
        self.top = top
        self.left = left
        self.bottom = top + height
        self.right = left + width
        self.height = height
        self.width = width
        self.scale = scale

    def crop(self, image):
        c = crop(image, self.left, self.top, self.width, self.height)
        return zoom(c, self.scale)

    @property
    def marker_pos(self):
        return self.get_marker_pos()

    def get_marker_pos(self):
        return [self.left, self.top]

    @property
    def marker_size(self):
        return self.get_marker_size()

    def get_marker_size(self):
        return [self.right - self.left, self.bottom - self.top]

class SplitImage:
    def __init__(self, list_img, vertical=True, degree=15, weights=None):
            '''
            This class allows to split several images and make one image out of them.
            The weights define how much space each image will take within that new image,
            and the degree along with vertical (boolean) decides how these images are split.
            Returns one raw image data.

            args:
                list_img (list of image arrays): provide preferably two or three images.
                    You can provide more (but it might look ugly).
                degree (integer): A value between -45° and 45°.
                vertical (boolean): Either uses a vertical or horizontal splitting.
                weights (list of floats): Matches the weights to each image in list_img.
            '''
            self.degree = degree
            if self.degree > 45 or self.degree < -45:
                print('Warning: SplitImage should get a degree between -45° and 45°, else the weights might not work like intended.' /
                'You might want to switch vertical from True to False or reverse.')
            self.is_vertical = vertical
            self.num_img = len(list_img)
            assert self.num_img > 1, "at least two images are required"

            self.weights = self._normalize_weights(weights)
            self.img_width = list_img[0].shape[1]
            self.img_height = list_img[0].shape[0]
            self.split_image = np.tile([x / 255 for x in [0,0,0]], (self.img_height, self.img_width, 1))
            # self.degree_rad = degree * np.pi / 180.0
            self.tan_degree_rad = np.tan(degree * np.pi / 180.0)

            self.start_pos = []
            self.end_pos = []
            self._make_split_image(list_img)

    def _calculate_default_weights(self):
        scale = 1.0 + int(self.degree / 10) * 0.25
        weights = np.zeros(self.num_img)
        weights.fill(scale)
        weights[0] = 1.0
        weights[-1] = 1.0
        return weights

    def _normalize_weights(self, weights):
        if weights is None:
            weights = self._calculate_default_weights()

        assert len(weights) == self.num_img, "need one weight per image"

        # normalize weight scaling
        weights /= np.sum(weights)
        return weights

    def _make_split_image(self, list_img):
        if not self.is_vertical:
            cur_pos = 0
            for i in range(len(list_img)):
                for col in range(self.img_width):
                    rel_w = float(self.img_width) * 0.5 - col
                    offset = self.tan_degree_rad * rel_w

                    if i == 0:
                        start = 0
                    else:
                        start = int(cur_pos + offset)

                    if i == len(list_img) - 1:
                        end = self.img_height
                    else:
                        end = int(cur_pos + self.weights[i] * self.img_height + offset)

                    start = max(0, start)
                    end = min(self.img_height, end)

                    # save start and end position to draw a line between images
                    self.split_image[start:end, col] = list_img[i][start:end, col]
                    if col == 0 and start != 0.:
                        self.start_pos.append((start, col))
                    if col == self.img_width-1 and end != self.img_height:
                        self.end_pos.append((end, col))

                cur_pos += self.weights[i] * self.img_height

        else:
            cur_pos = 0
            for i in range(len(list_img)):
                for row in range(self.img_height):
                    rel_h = float(self.img_height) * 0.5 - row
                    offset = self.tan_degree_rad * rel_h

                    if i == 0:
                        start = 0
                    else:
                        start = int(cur_pos + offset)

                    if i == len(list_img) - 1:
                        end = self.img_width
                    else:
                        end = int(cur_pos + self.weights[i] * self.img_width + offset)

                    start = max(0, start)
                    end = min(self.img_width, end)

                    self.split_image[row,start:end] = list_img[i][row,start:end]



                cur_pos += self.weights[i] * self.img_width

    def get_image(self):
        return self.split_image

    def get_start_positions(self):
        cur_pos = 0
        positions = []
        if not self.is_vertical:
            for i in range(1, self.num_img):
                cur_pos += self.weights[i-1] * self.img_height
                offset = self.tan_degree_rad * (float(self.img_width) * 0.5)
                r = int(cur_pos + offset)
                c = 0
                positions.append((r, c))
            return positions
        for i in range(1, self.num_img):
            cur_pos += self.weights[i-1] * self.img_width
            offset = self.tan_degree_rad * (float(self.img_height) * 0.5)
            c = int(cur_pos + offset)
            r = 0
            positions.append((r, c))
        return positions

    def get_end_positions(self):
        cur_pos = 0
        positions = []
        if not self.is_vertical:
            for i in range(1,self.num_img):
                cur_pos += self.weights[i-1] * self.img_height
                offset = self.tan_degree_rad * (float(self.img_width) * 0.5)
                r = int(cur_pos - offset)
                c = self.img_width
                positions.append((r, c))
            return positions
        for i in range(1,self.num_img):
            cur_pos += self.weights[i-1] * self.img_width
            offset = self.tan_degree_rad * (float(self.img_height) * 0.5)
            c = int(cur_pos - offset)
            r = self.img_height
            positions.append((r, c))
        return positions

    def get_weights(self):
        return self.weights

from simpleimageio import mse, relative_mse, relative_mse_outlier_rejection

def squared_error(img, ref):
    return (img - ref)**2

def relative_squared_error(img, ref, epsilon=0.0001):
    return (img - ref)**2 / (ref**2 + epsilon)

def sape(img, ref):
    ''' Computes the symmetric absolute precentage error
    '''
    err = np.absolute(ref - img)
    normalizer = np.absolute(img) + np.absolute(ref)
    mask = normalizer != 0
    err[mask] /= normalizer[mask]
    return err

def smape(img, ref):
    ''' Computes the symmetric mean absolute percentage error
    '''
    return np.average(sape(img,ref))



def outline(text: str, outline_clr=[10,10,10], text_clr=[250,250,250]) -> str:
    """ Creates a colored outline around the given text.

    Requires the "contour" package in the preamble, preferably with the [outline] option. That is:
    \usepackage[outline]{contour}

    Currently, this package is not included by default - might change in the future.

    Args:
        text: The text to outline
        outline_clr: Color of the outline as an sRGB triple
        text_clr: Color of the text as an sRGB triple

    Returns:
        LaTeX code to render the outlined text with the specified colors.
    """
    if outline_clr is None:
        res = "\\definecolor{FillClr}{RGB}{" + f"{text_clr[0]},{text_clr[1]},{text_clr[2]}" + "}"
        return res + "\\textcolor{FillClr}{" + text + "}"

    res = "\\DeclareDocumentCommand{\\Outlined}{ O{black} O{white} O{0.55pt} m }{"\
            "\\contourlength{#3}"\
            "\\contour{#2}{\\textcolor{#1}{#4}}"\
        "}"
    res += "\\definecolor{FillClr}{RGB}{" + f"{text_clr[0]},{text_clr[1]},{text_clr[2]}" + "}"
    res += "\\definecolor{StrokeClr}{RGB}{" + f"{outline_clr[0]},{outline_clr[1]},{outline_clr[2]}" + "}"

    res += "\\Outlined[FillClr][StrokeClr][0.55pt]{"+ text + "}"
    return res


from .element_data import *

import os
import tempfile
import subprocess
import shutil
import decimal

class PgfLinePlot(Plot):
    def __init__(self, aspect_ratio, data, dpi=300, axis_lines="left", tex_dir=None) -> None:
        """Creates a line plot using LaTeX with the pgfplots package

        Arguments:
            aspect_ratio (float): Height/width ratio of the plotting area (used for alignment and grid sizing)
            data (list): A list of plot lines. Each element is a pair of two equal-sized lists: the x and y coordinates.
            axis_lines (str): Either "left" (arrows on the left and bottom always) or "middle" (arrows at 0 coordinate)
        """
        self.aspect_ratio = aspect_ratio
        self._data = data
        self._markers = {}
        self.set_font(7, "{libertine}")
        self.set_linewidth(0.8, 0.6)
        self._colors = [
            [232, 181, 88],
            [5, 142, 78],
            [94, 163, 188],
            [181, 63, 106],
            [20, 20, 20]
        ]
        self._labels = {}
        self.set_axis_label("x", "", False)
        self.set_axis_label("y", "", True)
        self._axis_properties = {}
        self.set_axis_properties("x", [], use_log_scale=False)
        self.set_axis_properties("y", [], use_log_scale=False)
        self.set_padding(5, 5, 0)
        self._dpi = dpi
        self._axis_lines = axis_lines
        self._tex_dir = tex_dir
        self._legend = {}
        self.set_legend(line_length_mm=3, line_width_pt=1, pos=(1,1), anchor="north east",
            frame_line_width_pt=0.5, fontsize_pt=7, color_srgb=(100,100,100))

    def get_colors(self):
        return self._colors

    def set_colors(self, color_list):
        '''
        color list contains a list of colors. A color is defined as [r,g,b] while each channel
        ranges from 0 to 255.
        '''
        self._colors = color_list

    def set_legend(self, names=None, line_length_mm=None, line_width_pt=None, pos=None, anchor=None,
                   frame_line_width_pt=None, fontsize_pt=None, color_srgb=None):
        if names is not None:
            self._legend["names"] = names
        if line_length_mm is not None:
            self._legend["line_length"] = line_length_mm
        if line_width_pt is not None:
            self._legend["line_width"] = line_width_pt
        if pos is not None:
            self._legend["pos"] = pos
        if anchor is not None:
            self._legend["anchor"] = anchor
        if frame_line_width_pt is not None:
            self._legend["frame_line_width"] = frame_line_width_pt
        if fontsize_pt is not None:
            self._legend["fontsize"] = fontsize_pt
        if color_srgb is not None:
            self._legend["color"] = color_srgb

    def set_axis_label(self, axis, txt, vertical=False):
        self._labels[axis] = {}
        self._labels[axis]['text'] = txt.replace("\n", "\\\\{}")
        self._labels[axis]['vertical'] = vertical

    def set_axis_properties(self, axis, ticks, range=None, use_log_scale=True, use_scientific_notations=False):
        '''
        The user should find and define suitable ticks so that the labels and ticks don't overlap.
        Would be nice to do that automatically at some point.
        '''
        if range is not None and len(range) != 2:
            raise Error('You need exactly two values to specify range: [min, max]')

        self._axis_properties[axis] = {}
        if range is not None:
            self._axis_properties[axis]['range'] = range
        self._axis_properties[axis]['ticks'] = ticks
        self._axis_properties[axis]['use_log_scale'] = use_log_scale
        self._axis_properties[axis]['use_scientific_notations'] = use_scientific_notations

    def set_font(self, fontsize_pt=None, tex_package=None):
        if fontsize_pt is not None:
            self._fontsize_pt = fontsize_pt
        if tex_package is not None:
            self._font_tex_package = tex_package

    def set_linewidth(self, plot_line_pt=None, tick_line_pt=None):
        if plot_line_pt is not None:
            self._plot_linewidth_pt = plot_line_pt
        if tick_line_pt is not None:
            self._tick_linewidth_pt = tick_line_pt

    def set_padding(self, bottom_mm=None, left_mm=None, right_mm=None):
        if bottom_mm is not None:
            self._pad_bot_mm = bottom_mm
        if left_mm is not None:
            self._pad_left_mm = left_mm
        if right_mm is not None:
            self._pad_right_mm = right_mm

    def set_v_line(self, pos, color, linestyle=[], linewidth_pt=.8, phase_shift=0):
        ''' Adds a vertical line to the plot

        Args:
            color: the sRGB color of the line, each value in range [0, 255]
            linestyle: a list of dash lengths following the pattern: [on, off, on, off, ...].
                       An empty list corresponds to a solid line
            phase_shift: offset added to the dash pattern
        '''
        try:
            test = self._markers['vertical_lines'][0]
        except:
            self._markers['vertical_lines'] = []
        self._markers['vertical_lines'].append({
            'pos': pos,
            'color': color,
            "linestyle": linestyle,
            "linewidth_pt": linewidth_pt,
            "linephase": phase_shift,
        })

    def set_h_line(self, pos, color, linestyle=[], linewidth_pt=.8, phase_shift=0):
        ''' Adds a horizontal line to the plot
            Args:
                color: the sRGB color of the line, each value in range [0, 255]
                linestyle: a list of dash lengths following the pattern: [on, off, on, off, ...].
                           An empty list corresponds to a solid line
        '''
        try:
            test = self._markers['horizontal_lines'][0]
        except:
            self._markers['horizontal_lines'] = []
        self._markers['horizontal_lines'].append({
            'pos': pos,
            'color': color,
            "linestyle": linestyle,
            "linewidth_pt": linewidth_pt,
            "linephase": phase_shift,
        })

    def _compile_tex(self, tex, name):
        """ Compiles the given LaTeX code.

        Args:
            - tex   The file content of the LaTeX file to compile
            - name  Name of the output without the .pdf extension
            - intermediate_dir  Specify an existing directory here and .tex and .log files will be kept there
        """
        if self._tex_dir is not None and os.path.isdir(self._tex_dir):
            temp_folder = None
            temp_dir = os.path.abspath(self._tex_dir)
        else:
            temp_folder = tempfile.TemporaryDirectory()
            temp_dir = temp_folder.name

        with open(os.path.join(temp_dir, f"{os.path.basename(name)}.tex"), "w") as fp:
            fp.write(tex)

        subprocess.check_call(["pdflatex", "-interaction=nonstopmode", f"{os.path.basename(name)}.tex"],
            cwd=temp_dir, stdout=subprocess.DEVNULL)
        shutil.copy(os.path.join(temp_dir, f"{os.path.basename(name)}.pdf"), f"{name}.pdf")

        if temp_folder is not None:
            temp_folder.cleanup()

    def _ticks_to_str(self, axis):
        ticks = self._axis_properties[axis]['ticks']
        if ticks is None or len(ticks) == 0:
            return "\\empty"
        tick_str = [f"{t}" for t in ticks]
        return "{" + ",".join(tick_str) + "}"

    @staticmethod
    def _dash_pattern_to_str(pattern, phase):
        if pattern is None:
            return "{}"
        names = ["on", "off"]
        seq = "dash pattern = {"
        for i in range(len(pattern)):
            seq += f"{names[i % 2]} {pattern[i]} "
        seq += "},"
        phase = "dash phase = {" + str(phase) + "},"
        return seq + phase

    def _clip_ticks(self, axis):
        # Ensure that all ticks fall within the range, otherwise LaTeX will not compile
        if 'range' in self._axis_properties[axis]:
            if self._axis_properties[axis]['ticks'] is not None:
                clipped = []
                for t in self._axis_properties[axis]['ticks']:
                    if t > self._axis_properties[axis]['range'][0] and t < self._axis_properties[axis]['range'][1]:
                        clipped.append(t)
                self._axis_properties[axis]['ticks'] = clipped

    def _clean(self):
        self._clip_ticks("x")
        self._clip_ticks("y")

    def _make_tex(self, width, height):
        self._clean()

        tex_code = ""

        preamble_lines = [
            "\\documentclass{article}",
            "\\pagenumbering{gobble}",
            "\\usepackage{xcolor}",
            "\\usepackage{graphicx}",
            "\\usepackage[utf8]{inputenc}",
            "\\usepackage[T1]{fontenc}",
            "\\usepackage{geometry}",
            "\\usepackage{tikz}",
            "\\usepackage{pgfplots}",
            "\\pgfplotsset{compat=newest}",
            "\\pgfplotsset{",
            "    legend image code/.code={",
            f"        \\draw[mark repeat=2,mark phase=2,line width={self._legend['line_width']}pt]",
            f"        plot coordinates {{(0cm,0cm)({self._legend['line_length']}mm,0cm)}};",
            "    }",
            "}",

            "\\usepackage" + self._font_tex_package,

            "\\newcommand{\\width}{" + f"{width}mm" + "}",
            "\\newcommand{\\height}{" + f"{height}mm" + "}",
            "\\newcommand{\\padbot}{" + f"{self._pad_bot_mm}mm" + "}",
            "\\newcommand{\\padleft}{" + f"{self._pad_left_mm}mm" + "}",
            "\\newcommand{\\padright}{" + f"{self._pad_right_mm}mm" + "}",

            "\\geometry{",
            "    papersize={\\width,\\height},",
            "    total={\\width,\\height},",
            "    left=0mm,",
            "    top=0mm,",
            "}",

            "\\makeatletter \\newcommand{\\pgfplotsdrawaxis}{\\pgfplots@draw@axis} \\makeatother",
            "\\pgfplotsset{axis line on top/.style={",
            "axis line style=transparent,",
            "ticklabel style=transparent,",
            "tick style=transparent,",
            "axis on top=false,",
            "after end axis/.append code={",
            "    \\pgfplotsset{axis line style=opaque,",
            "    ticklabel style=opaque,",
            "    tick style=opaque,",
            "    grid=none}",
            "    \\pgfplotsdrawaxis}",
            "  }",
            "}",
            "\\definecolor{legendframecolor}{RGB}{" +
                f"{self._legend['color'][0]},{self._legend['color'][1]},{self._legend['color'][2]}" + "}",
        ]

        if self._colors is not None:
            i = 0
            for clr in self._colors:
                preamble_lines.append("\\definecolor{color" + f"{i}" + "}{RGB}{"
                    + f"{clr[0]},{clr[1]},{clr[2]}" + "}")
                i += 1

        if 'vertical_lines' in self._markers:
            i = 0
            for m in self._markers['vertical_lines']:
                clr = m["color"]
                preamble_lines.append("\\definecolor{vertlinecolor" + f"{i}" + "}{RGB}{"
                    + f"{clr[0]},{clr[1]},{clr[2]}" + "}")
                i += 1

        if 'horizontal_lines' in self._markers:
            i = 0
            for m in self._markers['horizontal_lines']:
                clr = m["color"]
                preamble_lines.append("\\definecolor{horzlinecolor" + f"{i}" + "}{RGB}{"
                    + f"{clr[0]},{clr[1]},{clr[2]}" + "}")
                i += 1
        preamble_lines.append("\\definecolor{gridcolor}{RGB}{220,220,220}")

        tex_code += "\n".join(preamble_lines) + "\n"

        body_start_lines = [
            "\\begin{document}",
            "\\raggedleft",
            "\\begin{tikzpicture}[trim axis left]",
            "\\clip (-\\padleft,-\\padbot) rectangle (\\width-\\padleft, \\height-\\padbot);",
            "\\begin{axis}[",
            "    scale only axis,",
            "    height=\\height-\\padbot,",
            "    width=\\width-\\padleft-\\padright,",
            f"    axis lines = {self._axis_lines},",
            "    xlabel near ticks,",
            "    xlabel={" + self._labels["x"]["text"] + "},",
            "    ylabel={" + self._labels["y"]["text"] + "},",
            f"    xtick={self._ticks_to_str('x')},",
            f"    ytick={self._ticks_to_str('y')},",
            "    yminorticks=false,",
            "    xminorticks=false,",
            "    yticklabel style={inner sep=1pt},", # distance between y tick labels and the marker
            "    xticklabel style={inner sep=1pt},", # distance between x tick labels and the marker
            "    legend pos=north west,",
            "    ymajorgrids=true,",
            "    xmajorgrids=true,",
            "    grid style={solid, line width=0.25pt, gridcolor},",
            "    axis line on top,",
            "    xlabel style={",
            "        inner sep=3pt,",
            "        at={([xshift=3pt] ticklabel* cs:1.0)}, anchor=north east,",
            "        text width=\\width,",
            "        align=right,",
            "    },",
            "    ylabel style={",
            "        inner sep=3pt,"

            "        at={([yshift=3pt] ticklabel* cs:1.0)}, anchor=south east, "
                if self._labels["y"]["vertical"] else
            "        at={([yshift=0pt] ticklabel* cs:1.0)}, anchor=north east,",

            "        rotate=0," if self._labels["y"]["vertical"] else "        rotate=270,",
            "        text width=\\height,",
            "        align=right,",
            "    },",
            "    label style={",
            "        font=\\fontsize{" + f"{self._fontsize_pt}" + "pt}{" + f"{self._fontsize_pt}" + "pt}\\selectfont",
            "    },",
            "    tick label style={",
            "        font=\\fontsize{" + f"{self._fontsize_pt}" + "pt}{" + f"{self._fontsize_pt}" + "pt}\\selectfont",
            "    },",
            f"    line width={self._plot_linewidth_pt}pt,",
            "    axis line style={line width=" + f"{self._tick_linewidth_pt}" + "pt},",
            "    tick style={",
            f"        line width={self._tick_linewidth_pt}pt,",
            "        color=black",
            "    },",
            "    x tick label style={",
            "        /pgf/number format/.cd,",
            "        scaled x ticks = false,",
            "        fixed," if not self._axis_properties["x"]['use_scientific_notations'] else "",
            "        /tikz/.cd",
            "    },",
            "    y tick label style={",
            "        /pgf/number format/.cd,",
            "        scaled y ticks = false,",
            "        fixed," if not self._axis_properties["y"]['use_scientific_notations'] else "",
            "        /tikz/.cd",
            "    },",
            "    legend cell align={left},",
            "    legend style = {",
            f"        at={{({self._legend['pos'][0]},{self._legend['pos'][1]})}},",
            f"        anchor ={self._legend['anchor']},",
            f"        line width={self._legend['frame_line_width']},",
            "        draw=legendframecolor,",
            f"        font=\\fontsize{{ {self._legend['fontsize']}pt}}{{ {self._legend['fontsize']}pt}}\\selectfont,",
            "    },",
        ]

        if not self._axis_properties["x"]['use_scientific_notations']:
            body_start_lines.append("    log ticks with fixed point,")

        if 'range' in self._axis_properties['x']:
            body_start_lines.append(f"    xmin={self._axis_properties['x']['range'][0]}, xmax={self._axis_properties['x']['range'][1]},")
            # body_start_lines.append(f"    restrict x to domain={self._axis_properties['x']['range'][0]}:{self._axis_properties['x']['range'][1]},")
        if 'range' in self._axis_properties['y']:
            body_start_lines.append(f"    ymin={self._axis_properties['y']['range'][0]}, ymax={self._axis_properties['y']['range'][1]},")
            # body_start_lines.append(f"    restrict y to domain={self._axis_properties['y']['range'][0]}:{self._axis_properties['y']['range'][1]},")

        if self._axis_properties["x"]['use_log_scale']:
            body_start_lines.append("    xmode=log,")
        if self._axis_properties["y"]['use_log_scale']:
            body_start_lines.append("    ymode=log,")

        body_start_lines.append("]")
        tex_code += "\n".join(body_start_lines) + "\n"

        # Add the actual plot lines
        for line_idx in range(len(self._data)):
            plot_code = [
                "\\addplot[",
                f"    color=color{line_idx}" if self._colors is not None and len(self._colors) > line_idx else "",
                "]",
                "coordinates {",
            ]
            coords = ""
            for i in range(len(self._data[line_idx][0])):
                x = self._data[line_idx][0][i]
                y = self._data[line_idx][1][i]
                # Format as fixed point because pgfplots does not understand scientific notation.
                # We convert to decimal, because that is the easiest way to get Python to include all digits
                coords += f"({decimal.Decimal(x):f},{decimal.Decimal(y):f})"
            plot_code.append(coords)
            plot_code.append("};")

            plot_code = "\n".join(plot_code) + "\n"

            # Long lines of text will cause pdflatex to crash, so we need to wrap the lines
            max_line_len = 200
            if len(plot_code) > max_line_len:
                while plot_code:
                    tex_code += plot_code[ : max_line_len] + "\n"
                    plot_code = plot_code[max_line_len : ]
            else:
                tex_code += plot_code

            if "names" in self._legend and self._legend["names"] is not None:
                tex_code += "\\addlegendentry{" + self._legend["names"][line_idx] + "}\n"

        # Add vertical and horizontal markers
        if "vertical_lines" in self._markers:
            i = 0
            for m in self._markers["vertical_lines"]:
                codelines = [
                    "\\draw[",
                    f"  vertlinecolor{i},",
                    f"  line width={m['linewidth_pt']}pt,",
                    self._dash_pattern_to_str(m["linestyle"], m["linephase"]),
                    "]",
                    "({axis cs:" + f"{m['pos']}" + ",0}|-{rel axis cs:0,1}) -- ({axis cs:"
                        + f"{m['pos']}" + ",0}|-{rel axis cs:0,0});"
                ]
                tex_code += "\n".join(codelines) + "\n"
                i += 1
        if "horizontal_lines" in self._markers:
            i = 0
            for m in self._markers["horizontal_lines"]:
                codelines = [
                    "\\draw[",
                    f"  horzlinecolor{i},",
                    f"  line width={m['linewidth_pt']}pt,",
                    self._dash_pattern_to_str(m["linestyle"], m["linephase"]),
                    "]",
                    "({rel axis cs:1,0}|-{axis cs:0," + f"{m['pos']}" +
                        "}) -- ({rel axis cs:0,0}|-{axis cs:0," + f"{m['pos']}" + "});"
                ]
                tex_code += "\n".join(codelines) + "\n"
                i += 1

        body_end_lines = [
            "\\end{axis}",
            "\\end{tikzpicture}",
            "\\end{document}",
        ]
        tex_code += "\n".join(body_end_lines) + "\n"

        return tex_code

    def make_pdf(self, width, height, filename):
        tex = self._make_tex(width, height)
        self._compile_tex(tex, filename)
        return filename + ".pdf"

    def make_raster(self, width, height, filename):
        fn = self.make_pdf(width, height, filename)
        from pdf2image import convert_from_path
        convert_from_path(fn, dpi=self._dpi, transparent=True, fmt="png", output_file=filename, single_file=True)
        return filename + ".png"


from pptx import Presentation
from pptx.util import Mm, Pt
from pptx.dml.color import RGBColor
from pptx.enum.shapes import MSO_SHAPE, MSO_CONNECTOR
from pptx.enum.text import PP_ALIGN
import importlib.resources as pkg_resources
from concurrent.futures import ThreadPoolExecutor, Future
from threading import Lock
from .backend import *

class PptxBackend(Backend):
    ''' A very basic .pptx generator that only roughly matches results of other backends.

    in PPTX format we have the following limitations due to our dependency python-pptx:
    - ignore background colors
    - ignore vertical alignment and padding of titles
    - do not support 'dashed' frames - if a frame is 'dashed' the frame in pptx will be normal (but still has a frame)
    - only support text rotation by 0° and +-90°
    '''
    def __init__(self):
        self._thread_pool = ThreadPoolExecutor()
        self._slide_mutex = Lock()

    def assemble_grid(self, components: List[Component], output_dir: str):
        return components

    def combine_grids(self, data: List[List[Component]], idx: int, bounds: Bounds) -> List[Component]:
        flat = []
        for row in data:
            flat.extend(row)
        return flat

    def _add_image(self, c: Component, slide):
        # Write image to temp folder
        with tempfile.TemporaryDirectory() as tmpdir:
            fname = c.data.make_raster(c.bounds.width, c.bounds.height, os.path.join(tmpdir, "image"))
            self._slide_mutex.acquire()
            shape = slide.shapes.add_picture(fname, Mm(c.bounds.left), Mm(c.bounds.top),
                width=Mm(c.bounds.width))
            shape.shadow.inherit = False
            self._slide_mutex.release()

        if c.has_frame:
            self._slide_mutex.acquire()
            offset = Pt(c.frame_linewidth) / 2
            shape = slide.shapes.add_shape(MSO_SHAPE.RECTANGLE, Mm(c.bounds.left) + offset,
                Mm(c.bounds.top) + offset, Mm(c.bounds.width) - Pt(c.frame_linewidth),
                Mm(c.bounds.height) - Pt(c.frame_linewidth))
            shape.shadow.inherit = False
            shape.line.color.rgb = RGBColor(c.frame_color[0], c.frame_color[1], c.frame_color[2])
            shape.line.width = Pt(c.frame_linewidth)
            # shape.line.join_type = 'Miter' # Removes rounded edges, but is not supported, yet (sadly)
            shape.fill.background()
            self._slide_mutex.release()

    def combine_rows(self, data: List[Component], bounds: Bounds):
        # We load a template from a file to have some nicer line styles etc by default
        # (they cannot currently be specified via python-pptx)
        with tempfile.TemporaryDirectory() as tmpdir:
            themedata = pkg_resources.read_binary(__package__, "theme.pptx")
            p = os.path.join(tmpdir, "theme.pptx")
            with open(p, "wb") as f:
                f.write(themedata)
            prs = Presentation(p)

        # Create a single slide presentation with a blank slide
        # prs = Presentation()
        prs.slide_height = Mm(bounds.height)
        prs.slide_width = Mm(bounds.width)
        blank_slide_layout = prs.slide_layouts[6]
        slide = prs.slides.add_slide(blank_slide_layout)

        # Add all our elements to the slide
        flat = []
        for row in data:
            flat.extend(row)

        # Generate all images in parallel
        futures = []
        for c in flat:
            if isinstance(c, ImageComponent):
                futures.append(self._thread_pool.submit(self._add_image, c, slide))
        for f in futures:
            f.result()

        # Add everything else afterwards, to ensure proper z-order
        for c in flat:
            if isinstance(c, TextComponent):
                if c.rotation == 90.0 or c.rotation == -90.0:
                    # The shape is rotated about its center. We want a rotation about the top left corner instead.
                    # Since we only allow 90° rotations, we can correct for that with a simple translation
                    pos_top = c.bounds.top + c.bounds.height / 2. - c.bounds.width / 2.
                    pos_left = c.bounds.left - c.bounds.height / 2. + c.bounds.width / 2.

                    # swap height and width
                    height, width = c.bounds.width, c.bounds.height

                    shape = slide.shapes.add_shape(MSO_SHAPE.RECTANGLE, Mm(pos_left), Mm(pos_top),
                        Mm(width), Mm(height))
                    # tikz rotation is counter-clockwise, pptx clockwise (we switch in pptx)
                    shape.rotation = -c.rotation
                else:
                    shape = slide.shapes.add_shape(MSO_SHAPE.RECTANGLE, Mm(c.bounds.left), Mm(c.bounds.top),
                        Mm(c.bounds.width), Mm(c.bounds.height))

                shape.shadow.inherit = False

                # Background color
                if c.background_color is not None:
                    shape.fill.solid()
                    shape.fill.fore_color.rgb = RGBColor(c.background_color[0], c.background_color[1],
                        c.background_color[2])
                else:
                    shape.fill.background()
                shape.line.fill.background()

                # Text properties
                text_frame = shape.text_frame
                p = text_frame.paragraphs[0]
                p.alignment = {
                    "center": PP_ALIGN.CENTER, "left": PP_ALIGN.LEFT, "right": PP_ALIGN.RIGHT
                }[c.horizontal_alignment]

                text_frame.margin_top = 0
                text_frame.margin_bottom = 0

                if c.horizontal_alignment == 'right':
                    text_frame.margin_right = Mm(c.padding.width_mm)
                    text_frame.margin_left = 0
                else:
                    text_frame.margin_right = 0
                    text_frame.margin_left = Mm(c.padding.width_mm)

                run = p.add_run()
                run.text = c.content.replace("\\\\", "\n")
                run.font.color.rgb = RGBColor(c.color[0], c.color[1], c.color[2])
                run.font.size = Pt(c.fontsize)

            if isinstance(c, RectangleComponent):
                shape = slide.shapes.add_shape(MSO_SHAPE.RECTANGLE, Mm(c.bounds.left), Mm(c.bounds.top),
                        Mm(c.bounds.width), Mm(c.bounds.height))
                shape.shadow.inherit = False
                shape.line.color.rgb = RGBColor(c.color[0], c.color[1], c.color[2])
                shape.line.width = Pt(c.linewidth)
                # shape.line.join_type = 'Miter' # Removes rounded edges, but is not supported, yet (sadly)
                shape.fill.background()

            if isinstance(c, LineComponent):
                shape = slide.shapes.add_connector(MSO_CONNECTOR.STRAIGHT, Mm(c.from_x), Mm(c.from_y),
                    Mm(c.to_x), Mm(c.to_y))
                shape.shadow.inherit = False
                shape.line.color.rgb = RGBColor(c.color[0], c.color[1], c.color[2])
                shape.line.width = Pt(c.linewidth)

        return prs

    def write_to_file(self, data, filename: str):
        data.save(filename)


from dataclasses import dataclass
from .figuregen import *

@dataclass
class Size:
    width_mm: float
    height_mm: float

def pt_to_mm(x):
    return x * 0.352778

def _get_space_type(position: str) -> str:
    if position == 'east' or position == 'west':
        return 'width'
    return 'height'

def sum_caption_spacing(layout: LayoutView, position: str, factor: float) -> float:
    sum = 0
    caption = layout.layout['element_config']['captions'][position]
    spacing = _get_space_type(position)
    if (caption[spacing] > 0.0):
        sum = caption['offset'] + caption[spacing]
    return sum * factor

def sum_title_spacing(layout: LayoutView, position):
    title = layout.layout['titles'][position]
    spacing = _get_space_type(position)
    if (title[spacing] > 0.0):
        return title['offset'] + title[spacing]
    return 0

def sum_row_title_spacing(layout: LayoutView, position):
    if (layout.layout['row_titles'][position]['width'] > 0.0):
        return layout.layout['row_titles'][position]['offset'] + layout.layout['row_titles'][position]['width']
    return 0

def sum_col_title_spacing(layout: LayoutView, position):
    if (layout.layout['column_titles'][position]['height'] > 0.0):
        return layout.layout['column_titles'][position]['height'] + layout.layout['column_titles'][position]['offset']
    return 0

def min_width(grid: Grid):
    '''
    Minimum width is the sum of all fixed space/padding based on the json-config file, including titles, offsets, paddings, etc.
    So basically: everything except for the images.
    '''
    min_width = grid.layout.layout['column_space'] * (grid.cols - 1)# + (muliplied by num -1)
    min_width += grid.layout.layout['padding']['west']
    min_width += grid.layout.layout['padding']['east']
    min_width += sum_caption_spacing(grid.layout, 'east', grid.cols)
    min_width += sum_caption_spacing(grid.layout, 'west', grid.cols)
    min_width += sum_title_spacing(grid.layout, 'east')
    min_width += sum_title_spacing(grid.layout, 'west')
    min_width += sum_row_title_spacing(grid.layout, 'east')
    min_width += sum_row_title_spacing(grid.layout, 'west')
    return min_width

def fixed_inner_width(grid: Grid):
    '''
    Fixed inner width is the sum of spacing between all images, which also includes element captions
    '''
    inner_width = grid.layout.layout['column_space'] * (grid.cols - 1)
    inner_width += sum_caption_spacing(grid.layout,'east', grid.cols - 1)
    inner_width += sum_caption_spacing(grid.layout,'west', grid.cols - 1)
    return inner_width

def body_width(grid: Grid, image_size: Size):
    '''
    body means: all images and their spaces/padding inbetween the images.
    Not included are: column/row titles and titles as well as their corresping offsets.
    '''
    return fixed_inner_width(grid) + grid.cols * image_size.width_mm

def total_width(grid: Grid, image_size: Size):
    '''
    Includes everything that takes up width: padding, images, captions, row titles,
    east/west titles and all corresponding offsets
    '''
    total_width = body_width(grid, image_size)
    total_width += sum_caption_spacing(grid.layout, 'east', 1) # add to inner body one more
    total_width += sum_caption_spacing(grid.layout, 'west', 1) # same reason as above

    # add outer titles
    total_width += sum_row_title_spacing(grid.layout, 'east')
    total_width += sum_row_title_spacing(grid.layout, 'west')
    total_width += sum_title_spacing(grid.layout, 'east')
    total_width += sum_title_spacing(grid.layout, 'west')

    total_width += grid.layout.layout['padding']['west'] + grid.layout.layout['padding']['east']
    return total_width

def min_height(grid: Grid):
    '''
    Minimum height is the sum of all fixed space/padding based on the json-config file, including titles, offsets, paddings, etc.
    So basically: everything except for the images.
    '''
    min_height = grid.layout.layout['row_space'] * (grid.rows -1)
    min_height += grid.layout.layout['padding']['north']
    min_height += grid.layout.layout['padding']['south']
    min_height += sum_caption_spacing(grid.layout, 'north', grid.rows)
    min_height += sum_caption_spacing(grid.layout, 'south', grid.rows)
    min_height += sum_title_spacing(grid.layout, 'north')
    min_height += sum_title_spacing(grid.layout, 'south')
    min_height += sum_col_title_spacing(grid.layout, 'north')
    min_height += sum_col_title_spacing(grid.layout, 'south')
    return min_height

def fixed_inner_height(grid: Grid):
    '''
    Fixed inner height is the sum of spacing between all images, which also includes element captions
    '''
    inner_height = grid.layout.layout['row_space'] * (grid.rows -1)
    inner_height += sum_caption_spacing(grid.layout, 'north', grid.rows-1)
    inner_height += sum_caption_spacing(grid.layout, 'south', grid.rows-1)

    return inner_height

def body_height(grid: Grid, image_size: Size):
    '''
    body means: all images and their spaces/padding inbetween the images.
    Not included are: column/row titles and titles as well as their corresping offsets.
    '''
    return fixed_inner_height(grid) + grid.rows * image_size.height_mm

def total_height(grid: Grid, image_size: Size):
    '''
    Includes everything that takes up height: padding, images, captions, column titles,
    north/south titles and all corresponding offsets
    '''
    total_height = body_height(grid, image_size)
    total_height += sum_caption_spacing(grid.layout, 'north', 1) # add to inner body one more
    total_height += sum_caption_spacing(grid.layout, 'south', 1) # add to inner body one more

    # add outer titles
    total_height += sum_col_title_spacing(grid.layout, 'north')
    total_height += sum_col_title_spacing(grid.layout, 'south')
    total_height += sum_title_spacing(grid.layout, 'north')
    total_height += sum_title_spacing(grid.layout, 'south')

    total_height += grid.layout.layout['padding']['north'] + grid.layout.layout['padding']['south']
    return total_height

def element_size_from_width(grid: Grid, total_width: float) -> Size:
    """ Computes the size of all individual images in the grid based on the given total width. """
    min_w = min_width(grid)
    width_per_img = (total_width - min_w) / grid.cols
    if width_per_img < 1.0:
        if width_per_img < 0.0:
            print(f'Warning: Element width computed to be negative. Probably due to an extreme aspect ratio.'
                  f'Total height: ({total_width} - {min_w}) / {grid.cols} = {width_per_img}')
        else:
            print(f'Warning: Width per element is {width_per_img}, which is less than 1mm.'
                   'Probably due to an extreme aspect ratio or too many elements.')

    return Size(width_per_img, width_per_img * grid.aspect_ratio)

def element_size_from_height(grid: Grid, total_height: float) -> Size:
    """ Computes the size of all individual images in the grid based on the given total height. """
    min_h = min_height(grid)
    height_per_img = (total_height - min_h) / grid.rows
    if height_per_img < 1.0:
        if height_per_img < 0.0:
            print(f'Warning: Element height computed to be negative. Probably due to an extreme aspect ratio.'
                  f'Total height: ({total_height} - {min_h}) / {grid.rows} = {height_per_img}')
        else:
            print(f'Warning: Height per element is {height_per_img} which is less than 1mm.'
                   'Probably due to an extreme aspect ratio or too many elements.')
    return Size(height_per_img / grid.aspect_ratio, height_per_img)

def size_of(data_part, direction: str):
    assert direction in data_part

    def _get_size(space, offset):
        if space == 0.0:
            return 0.0, 0.0
        return space, offset

    if direction == 'north' or direction == 'south':
        return _get_size(data_part[direction]['height'], data_part[direction]['offset'])
    elif direction == 'east' or direction == 'west':
        return _get_size(data_part[direction]['width'], data_part[direction]['offset'])
    else:
        raise Error("Error: Invalid direction value: " + direction +". (html module)")

def image_pos(grid: Grid, image_size: Size, column, row):
    """ Computes the position of an image in the given grid, relative to the grid's top left corner. """
    title_top = sum(size_of(grid.layout.layout['titles'], 'north'))
    title_left = sum(size_of(grid.layout.layout['titles'], 'west'))
    col_title_top = sum(size_of(grid.layout.layout['column_titles'], 'north'))
    row_title_left = sum(size_of(grid.layout.layout['row_titles'], 'west'))
    img_south_capt = sum(size_of(grid.layout.layout['element_config']['captions'], 'south')) * (row)

    top = grid.layout.layout['padding']['north'] + title_top + col_title_top
    top += (grid.layout.layout['row_space'] + image_size.height_mm)*(row) + img_south_capt

    left =  grid.layout.layout['padding']['west'] + title_left + row_title_left
    left += (grid.layout.layout['column_space'] + image_size.width_mm)*(column)

    return top, left

def south_caption_pos(grid: Grid, image_size: Size, column, row):
    top, left = image_pos(grid, image_size, column, row)
    top += image_size.height_mm + grid.layout.layout['element_config']['captions']['south']['offset']
    return top, left

def titles_pos_and_size(grid: Grid, image_size: Size, direction: str):
    '''
    Note: this does not include element captions, yet. Because it was never really used in tikz or elsewhere.
    '''
    offset_left = grid.layout.layout['padding']['west']
    offset_top =  grid.layout.layout['padding']['north']

    if direction == 'north' or direction == 'south':
        width = image_size.width_mm * grid.cols + grid.layout.layout['column_space']*(grid.cols - 1)
        height = size_of(grid.layout.layout['titles'], direction)[0]

        offset_left += sum(size_of(grid.layout.layout['titles'], 'west')) + sum(size_of(grid.layout.layout['row_titles'], 'west'))
        if direction == 'south':
            offset_top += sum(size_of(grid.layout.layout['element_config']['captions'], 'south')) * grid.rows
            offset_top += grid.layout.layout['padding']['north'] + sum(size_of(grid.layout.layout['titles'], 'north')) + sum(size_of(grid.layout.layout['column_titles'], 'north'))
            offset_top += (grid.layout.layout['row_space']*(grid.rows - 1)) + image_size.height_mm * grid.rows
            offset_top += sum(size_of(grid.layout.layout['column_titles'], 'south')) + size_of(grid.layout.layout['titles'], 'south')[1]

        return offset_top, offset_left, width, height

    elif direction == 'east' or direction == 'west':
        height = image_size.height_mm * grid.rows + grid.layout.layout['row_space']*(grid.rows - 1)
        height += sum(size_of(grid.layout.layout['element_config']['captions'], 'south')) * (grid.rows - 1)
        width = size_of(grid.layout.layout['titles'], direction)[0]

        offset_top += sum(size_of(grid.layout.layout['titles'], 'north'))
        offset_top += sum(size_of(grid.layout.layout['column_titles'], 'north'))
        if direction == 'east':
            offset_left += sum(size_of(grid.layout.layout['titles'], 'west')) + sum(size_of(grid.layout.layout['row_titles'], 'west'))
            offset_left += (grid.layout.layout['column_space']*(grid.cols - 1)) + image_size.width_mm * grid.cols
            offset_left += sum(size_of(grid.layout.layout['row_titles'], 'east')) + size_of(grid.layout.layout['titles'], 'east')[1]

        return offset_top, offset_left, width, height

    else:
        assert False, "Error: Invalid direction value: " + direction +". (html module)"

def row_titles_pos(grid: Grid, image_size: Size, cur_row, direction):
    if not(direction == 'east' or direction == 'west'):
        raise Error("Error: Invalid direction value for row titles: " + direction +". Expected 'east' or 'west'.")

    data = grid.layout.layout

    width = size_of(data['row_titles'], direction)[0]
    height = image_size.height_mm

    offset_left = data['padding']['west'] + sum(size_of(data['titles'], 'west'))
    offset_top = data['padding']['north'] + sum(size_of(data['titles'], 'north')) + sum(size_of(data['column_titles'], 'north'))
    offset_top += (data['row_space'] + image_size.height_mm) * (cur_row - 1)
    offset_top += sum(size_of(data['element_config']['captions'], 'south')) * (cur_row-1)
    if direction == 'east':
        offset_left += sum(size_of(data['row_titles'], 'west'))
        offset_left += (data['column_space']*(grid.cols - 1)) + image_size.width_mm * grid.cols
        offset_left += size_of(data['row_titles'], 'east')[1]

    return offset_top, offset_left, width, height

def column_titles_pos(grid: Grid, image_size: Size, cur_column, direction):
    if not(direction == 'north' or direction == 'south'):
        raise "Error: Invalid direction value for column titles: " + direction +". Expected 'north' or 'south'."

    data = grid.layout.layout

    width = image_size.width_mm
    height = size_of(data['column_titles'], direction)[0]

    offset_top = data['padding']['north'] + sum(size_of(data['titles'], 'north'))
    offset_left = data['padding']['west'] + sum(size_of(data['titles'], 'west'))
    offset_left += (data['column_space'] + image_size.width_mm) *(cur_column - 1)
    offset_left += sum(size_of(data['row_titles'], 'west'))
    if direction == 'south':
        offset_top += sum(size_of(data['element_config']['captions'], 'south')) * grid.rows
        offset_top += sum(size_of(data['column_titles'], 'north'))
        offset_top += (data['row_space']*(grid.rows - 1)) + image_size.height_mm * grid.rows
        offset_top += size_of(data['column_titles'], 'south')[1]

    return offset_top, offset_left, width, height


from .element_data import *
from .util import units

import numpy as np
from threading import Lock

import matplotlib
matplotlib.use('pgf')
import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter

matplot_mutex = Lock()

# ------ HELPER -------
def _check_axis(axis):
    if not (axis in ['x', 'y']):
        raise Error('Incorrect axis. Try: "x" or "y".')

def _default_rotation(axis):
    if axis == 'x':
        return 'horizontal'
    return 'vertical'

def _interpret_rotation(rotation):
    if rotation == 0:
        return 'horizontal'
    if rotation == 90 or rotation == -90:
        return 'vertical'
    if rotation in ['horizontal', 'vertical']:
        return rotation
    raise Error('Incorrect rotation value. Try: 0/(-)90 or "horizontal"/"vertical".')

def _setup_fonts(plt, font_properties):
    plt.rcParams.update({
    "text.usetex": True,     # use inline math for tikz
    "pgf.rcfonts": False,    # don't setup fonts from rc parameters
    "pgf.texsystem": "pdflatex",
    "font.family": font_properties['font_family'],
    "pgf.preamble": "\n".join([
            r"\usepackage[utf8]{inputenc}",
            r"\usepackage[T1]{fontenc}",
            r"\usepackage" + font_properties['tex_package']
        ])
    })

def _label_alignment(rotation : str):
    return 'top' if rotation == 'horizontal' else 'bottom'

def _plot_lines(ax, data, colors, linewidth):
    i = 0
    for d in data:
        if colors is not None and len(colors) > i:
            ax.plot(d[0], d[1], linewidth=linewidth, color=np.array(colors[i])/255.0)
        else:
            ax.plot(d[0], d[1], linewidth=linewidth)
        i += 1

def _apply_axis_range(ax, props):
    if "range" in props["x"]:
        ax.set_xlim([ props["x"]['range'][0], props["x"]['range'][1] ])
    if "range" in props["y"]:
        ax.set_ylim([ props["y"]['range'][0], props["y"]['range'][1] ])

def _apply_axes_properties(ax, props):
    if props["x"]["use_log_scale"]:
        ax.set_xscale('log')
    if props["y"]["use_log_scale"]:
        ax.set_yscale('log')

    def set_ticks(ax, props):
        if props['ticks'] is not None:
            ax.set_ticks(props['ticks'])
            if not props['use_scientific_notations']: # can only apply if we have specific ticks
                ax.set_ticklabels(props['ticks'])
        ax.set_minor_formatter(FormatStrFormatter(""))

    set_ticks(ax.xaxis, props["x"])
    set_ticks(ax.yaxis, props["y"])

def _set_labels(fig, ax, labels, fontsize, pad):
    '''
    Sets fontsize (pt), labels and their rotation.
    The labels are placed at each end of the axes so that we don't waste too much space.
    The correct label position will be calculated automatically.
    Currently the user needs to find suitable ticks, so that labels and ticks don't overlap!
    '''
    axis_labels = labels
    ax.set_xlabel(axis_labels['x']['text'], fontsize=fontsize, ha="right",
                  va=_label_alignment(axis_labels['x']['rotation']), rotation=axis_labels['x']['rotation'])
    ax.set_ylabel(axis_labels['y']['text'], fontsize=fontsize, ha="right",
                  va=_label_alignment(axis_labels['y']['rotation']), rotation=axis_labels['y']['rotation'])

    # compute axis size in points
    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    width, height = bbox.width * 72, bbox.height * 72

    # compute relative padding
    lwX = ax.spines['bottom'].get_linewidth()
    # TODO if use_scientific_notations is True, then take ~70% of fontsize * ~1/4 and add it to the padding

    lwY = 0 #ax.spines['left'].get_linewidth()

    # coordinates in percentage of the figure main body size!
    ax.xaxis.set_label_coords(1, -(pad - lwX) / height)

    # coordinates in percentage of the figure main body size!
    ax.yaxis.set_label_coords(-(pad - lwY) / width, 1)

def _apply_axes_properties_and_labels(fig, ax, axis_properties, labels, config, fontsize):
    _apply_axis_range(ax, axis_properties)

    if not config['has_right_axis']:
        ax.spines['right'].set_visible(False)
        ax.yaxis.set_ticks_position('left')
    if not config['has_upper_axis']:
        ax.spines['top'].set_visible(False)
        ax.xaxis.set_ticks_position('bottom')

    tick_lw_pt = config['tick_linewidth_pt']
    plt.tick_params(width=tick_lw_pt, length=(tick_lw_pt * 4), labelsize=fontsize, pad=(tick_lw_pt * 2))
    _set_labels(fig, ax, labels, fontsize, pad=(tick_lw_pt * 6))
    # if use_scientific_notations True, displaystyle is used in pgf --> offset of ticks changes

    _apply_axes_properties(ax, axis_properties)

def _place_marker(ax, marker_data):
    try:
        vlines = marker_data['vertical_lines']
    except:
        vlines = []

    for vl in vlines:
        ax.axvline(x=vl['pos'], color=np.array(vl['color'])/255.0, linewidth=vl['linewidth_pt'],
            linestyle=vl['linestyle'])


# ------ FINALLY --------
class MatplotLinePlot(Plot):
    def __init__(self, aspect_ratio, data) -> None:
        """Creates a line plot using matplotlib with a pgf (i.e. LaTeX) backend

        Args:
            aspect_ratio (float): Height/width ratio of the plotting area (used for alignment and grid sizing)
            data (list): A list of plot lines. Each element is a pair of two equal-sized lists: the x and y coordinates.
        """
        self.aspect_ratio = aspect_ratio
        self._data = data
        self._labels = {}
        self._axis_properties = {}
        self._markers = {}
        self._font = {
            "tex_package": "{libertine}",
            "font_family": "sans-serif",
            "fontsize_pt": 7
        }
        self._grid = {
            "color": [ 230, 230, 230 ],
            "linewidth_pt": 0.25,
            "linestyle": "-"
        }
        self._config = {
            "plot_linewidth_pt": 0.8,
            "tick_linewidth_pt": 0.6,
            "has_upper_axis": False,
            "has_right_axis": False
        }
        self._colors = [
            [232, 181, 88],
            [5, 142, 78],
            [94, 163, 188],
            [181, 63, 106],
            [20, 20, 20]
        ]
        self.set_axis_label("x", "")
        self.set_axis_label("y", "")
        self.set_axis_properties("x", [], use_log_scale=False)
        self.set_axis_properties("y", [], use_log_scale=False)

    def get_colors(self):
        return self._colors

    def set_colors(self, color_list):
        '''
        color list contains a list of colors. A color is defined as [r,g,b] while each channel
        ranges from 0 to 255.
        '''
        self._colors = color_list

    def get_axis_label(self, axis=None):
        if axis is None:
            return self._labels
        _check_axis(axis)
        try:
            return self._labels[axis]
        except:
            raise Error('Label is not defined.')

    def set_axis_label(self, axis, txt, rotation=None):
        _check_axis(axis)

        if rotation is not None:
            rotation = _interpret_rotation(rotation)
        else:
            rotation = _default_rotation(axis)

        self._labels[axis] = {}
        self._labels[axis]['text'] = txt
        self._labels[axis]['rotation'] = rotation

    def get_axis_properties(self, axis):
        _check_axis(axis)
        try:
            return self._axis_properties[axis]
        except:
            raise Error('Label is not defined.')

    def set_axis_properties(self, axis, ticks, range=None, use_log_scale=True, use_scientific_notations=False):
        '''
        The user should find and define suitable ticks so that the labels and ticks don't overlap.
        Would be nice to do that automatically at some point.
        '''
        _check_axis(axis)
        if range is not None and len(range) != 2:
            raise Error('You need exactly two values to specify range: [min, max]')

        self._axis_properties[axis] = {}
        if range is not None:
            self._axis_properties[axis]['range'] = range
        self._axis_properties[axis]['ticks'] = ticks
        self._axis_properties[axis]['use_log_scale'] = use_log_scale
        self._axis_properties[axis]['use_scientific_notations'] = use_scientific_notations

    def get_v_line(self):
        try:
            test = self._markers['vertical_lines'][0]
            return self._markers['vertical_lines']
        except:
            return []

    def set_v_line(self, pos, color, linestyle, linewidth_pt=.8):
        '''
        Currently, we only implemented "vertical_lines"
        linestyle allows matplotlib inputs, e.g. (0,(4,6)) is valid.
        '''
        try:
            test = self._markers['vertical_lines'][0]
        except:
            self._markers['vertical_lines'] = []
        self._markers['vertical_lines'].append({
            'pos': pos,
            'color': color,
            "linestyle": linestyle,
            "linewidth_pt": linewidth_pt,
        })

    def get_font(self):
        return self._font

    def set_font(self, fontsize_pt=None, font_family=None, tex_package=None):
        if fontsize_pt is not None:
            self._font["fontsize_pt"] = fontsize_pt
        if font_family is not None:
            self._font["font_family"] = font_family
        if tex_package is not None:
            self._font["tex_package"] = tex_package

    def set_grid_properties(self, color=None, linewidth_pt=None, linestyle=None):
        if color is not None:
            self._grid["color"] = color
        if linewidth_pt is not None:
            self._grid["linewidth_pt"] = linewidth_pt
        if linestyle is not None:
            self._grid["linestyle"] = linestyle

    def show_upper_axis(self, show=True):
        self._config["has_upper_axis"] = show

    def show_right_axis(self, show=True):
        self._config["has_right_axis"] = show

    def set_linewidth(self, plot_line_pt=None, tick_line_pt=None):
        if plot_line_pt is not None:
            self._config['plot_linewidth_pt'] = plot_line_pt
        if tick_line_pt is not None:
            self._config['tick_linewidth_pt'] = tick_line_pt

    def _make(self, width_mm, height_mm, filename):
        matplot_mutex.acquire()
        try:
            _setup_fonts(plt, self._font)
            figsize = units.mm_to_inches(np.array([width_mm, height_mm]))

            #constrained_layout: https://matplotlib.org/3.2.1/tutorials/intermediate/constrainedlayout_guide.html
            fig, ax = plt.subplots(figsize=figsize, constrained_layout=True)
            fig.set_constrained_layout_pads(w_pad=0, h_pad=0, hspace=0., wspace=0.)

            _plot_lines(ax, self._data, self._colors, self._config['plot_linewidth_pt'])
            _apply_axes_properties_and_labels(fig, ax, self._axis_properties, self._labels,
                self._config, self._font['fontsize_pt'])
            plt.grid(color=np.array(self._grid['color'])/255.0, linestyle=self._grid['linestyle'],
                linewidth=self._grid['linewidth_pt'])
            _place_marker(ax, self._markers)
            plt.savefig(filename, pad_inches=0.0, dpi=500)
        finally:
            matplot_mutex.release()

    def make_raster(self, width_mm, height_mm, filename):
        self._make(width_mm, height_mm, filename + ".png")
        return filename + ".png"

    def make_pdf(self, width_mm, height_mm, filename):
        self._make(width_mm, height_mm, filename + ".pdf")
        return filename + ".pdf"


from typing import Union
from .backend import *
import importlib.resources as pkg_resources
from concurrent.futures import ThreadPoolExecutor, Future

class TikzBackend(Backend):
    """ Generates the code for a TikZ picture representing the figure.
    Default file ending is .tikz, use \\input{figure.tikz} to include in LaTeX.
    """

    def __init__(self, include_header=True):
        """
        Args:
            include_header: boolean, set to false to not include the macro definitions in the generated file.
        """
        self._include_header = include_header
        self._thread_pool = ThreadPoolExecutor()

    def add_overlay(self, tikz_code: str):
        """ Adds overlay code that will be stitched on top of the generated figure. """
        pass

    @property
    def header(self) -> str:
        """ The TikZ macros used for the figure components """
        return pkg_resources.read_text(__package__, 'commands.tikz')

    @property
    def preamble(self) -> str:
        """ The minimum set of \\usepackage's for the figure to display correctly. """
        return '\n'.join(["\\usepackage{calc}", "\\usepackage{tikz}"])

    def _sanitize_latex_path(self, path):
        # Assume that pdflatex will be run from the same folder and strip the directory name from the path
        return "\\detokenize{" + os.path.basename(path) + "}"

    def _latex_color(self, rgb):
        if rgb is None:
            return ""
        return "rgb,255:red," + str(rgb[0]) + ";green," + str(rgb[1]) + ";blue," + str(rgb[2])

    def _make_image(self, c: ImageComponent, dims: str, anchor: str, output_dir: str, elem_id: str) -> str:
        # Generate the image data
        prefix = "img-" + elem_id
        file_prefix = os.path.join(output_dir, prefix)
        try:
            filename = c.data.make_pdf(c.bounds.width, c.bounds.height, file_prefix)
        except NotImplementedError:
            filename = c.data.make_raster(c.bounds.width, c.bounds.height, file_prefix)

        # Assemble the position arguments
        fname = "{" + self._sanitize_latex_path(filename) + "}"
        name = "{" + prefix + "}"

        # Check if there is a frame and emit the correct command
        if c.has_frame:
            linewidth = "{" + f'{c.frame_linewidth}pt' + "}"
            color = "{" + self._latex_color(c.frame_color) + "}"
            return "\\makeframedimagenode" + dims + fname + name + anchor + color + linewidth + "\n"
        else:
            return "\\makeimagenode" + dims + fname + name + anchor + "\n"

    def assemble_grid(self, components: List[Component], output_dir: str) -> List[Union[Future, str]]:
        tikz_lines = []
        for c in components:
            elem_id = f"fig{c.figure_idx}-grid{c.grid_idx}"
            if c.row_idx >= 0:
                elem_id += f"-row{c.row_idx}"
            if c.col_idx >= 0:
                elem_id += f"-col{c.col_idx}"

            # Position arguments are the same for all components
            if c.bounds is not None:
                dims = "{" + f"{c.bounds.width}" + "mm}" + "{" + f"{c.bounds.height}" + "mm}"
                anchor = "{(" + f"{c.bounds.left}mm, {-c.bounds.top}mm" + ")}"

            if isinstance(c, ImageComponent):
                tikz_lines.append(self._thread_pool.submit(
                    self._make_image, c, dims, anchor, output_dir, elem_id))

            if isinstance(c, TextComponent):
                prefix = c.type + "-" + elem_id
                name = "{" + prefix + "}"
                fontsize = "{" + f'{c.fontsize}pt' + "}"
                color = "{" + self._latex_color(c.color) + "}"
                content = "{" + c.content + "}"
                rotation = "{" + str(c.rotation) + "}"
                fill_color = "{" + self._latex_color(c.background_color) + "}"

                node = "\\maketextnode" if c.rotation % 180 < 20 else "\\maketextnodeflipped"

                vert_align = "{c}"
                if c.vertical_alignment == "top":
                    vert_align = "{t}"
                elif c.vertical_alignment == "bottom":
                    vert_align = "{b}"

                horz_align = "{\\centering}"
                if c.horizontal_alignment == "left":
                    horz_align = "{\\raggedright}"
                elif c.horizontal_alignment == "right":
                    horz_align = "{\\raggedleft}"

                pad_vert = "{" + str(c.padding.height_mm) + "mm}"
                pad_horz = "{" + str(c.padding.width_mm) + "mm}"

                tikz_lines.append(node + dims + name + anchor + color + fontsize + fill_color + rotation +
                    vert_align + horz_align + pad_vert + pad_horz + content)

            if isinstance(c, RectangleComponent):
                color = "{" + self._latex_color(c.color) + "}"
                linewidth = "{" + str(c.linewidth) + "pt}"
                linestyle = "{dashed}" if c.dashed else "{solid}"
                tikz_lines.append("\\makerectangle" + dims + anchor + color + linewidth + linestyle)

            if isinstance(c, LineComponent):
                parent_name = "{" + "img-" + elem_id + "}"
                color = "{" + self._latex_color(c.color) + "}"
                linewidth = "{" + str(c.linewidth) + "pt}"
                start = "{" + f"({c.from_x}mm, {-c.from_y}mm)"+ "}"
                end = "{" + f"({c.to_x}mm, {-c.to_y}mm)"+ "}"
                tikz_lines.append("\\makeclippedline" + parent_name + start + end + linewidth + color)

        return tikz_lines

    def combine_grids(self, data, idx: int, bounds: Bounds) -> List[Union[Future, str]]:
        # Create an empty "background" node to make sure that outer paddings are not cropped away
        figure_id = "{figure-" + str(idx) + "}"
        dims = "{" + f"{bounds.width}" + "mm}" + "{" + f"{bounds.height}" + "mm}"
        anchor = "{(" + f"{bounds.left}mm, {-bounds.top}mm" + ")}"
        tikz_code = "\\makebackgroundnode" + dims + anchor + figure_id + "\n"

        # Flatten the list of list into a single list with one element per component
        result = [tikz_code]
        for grid in data:
            for d in grid:
                result.append(d)
        return result

    def combine_rows(self, data: List[List[Union[Future, str]]], bounds: Bounds) -> str:
        # Synchronize all export futures and combine the lines
        tikz_code = ""
        for fig in data:
            for c in fig:
                if isinstance(c, Future):
                    tikz_code += c.result() + "\n"
                else:
                    tikz_code += c + "\n"
        return tikz_code

    def write_to_file(self, data: str, filename: str):
        with open(filename, "w") as f:
            if self._include_header:
                f.write(self.header)
                f.write("\n")
            f.write("\\begin{tikzpicture}\n")
            f.write(data)
            f.write("\\end{tikzpicture}")


import setuptools

setuptools.setup(
    name='figuregen',
    version='1.0.0',
    description='Figure Generator',
    long_description='This tool generates figures in pdf- (via LaTeX), html- and pptx-format. '\
    'It might help not only to create final figures, but also to analyze images faster: '\
    'We offer a bunch of error metrics that allows not only to compare images visually but also mathematically. '\
    'The pptx-format allows manual tweaks - if so desired.',
    url='https://github.com/Mira-13/figure-gen',
    author='Mira Niemann',
    author_email='mira.niemann@gmail.com',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'matplotlib>=3.2.1',
        'python-pptx',
        'simpleimageio',
        'opencv-python'
    ],
    zip_safe=False,
    include_package_data=True
)



#PDF imports
from pdf2image import convert_from_path
import IPython
from IPython.display import Image
from IPython.display import display
import cv2
import numpy as np

# HTML imports
from IPython.core.display import HTML
import re

def loadpdf(pdfname, dpi=1000):
    images = convert_from_path(pdfname, dpi=dpi)
    return np.array(images[0])

def convert(pdfname, dpi=1000):
    img = loadpdf(pdfname, dpi)
    cv2.imwrite(pdfname.replace('.pdf', '.png'), cv2.cvtColor(img, cv2.COLOR_RGB2BGR))
    return pdfname.replace('.pdf', '.png')

def displaypdf(pdfname):
    filename = convert(pdfname)
    IPython.display.display(Image(filename))

def loadhtml(html_file):
    with open(html_file) as f:
        html = f.read()
        figure = (re.findall(r"<body>(.*)</body>", html, re.DOTALL)[0])

        table = {
            "module": "position: absolute; ",
            "title-container": "position: absolute; margin-top: 0; margin-bottom: 0;display: flex; align-items: center; justify-content: center; ",
            "title-content": "margin-top: 0; margin-bottom: 0;",
            "element": "position: absolute; margin: 0; "
        }

        for c, s in table.items():
            figure = figure.replace(f'class="{c}" style="', f'class="{c}" style="{s}')
            figure = figure.replace(f'class="{c}"', '')

        return figure

def displayhtml(html_file):
    display(HTML(data=loadhtml(html_file)))


from typing import List
from .default_layouts import layouts as _default_layout
from .element_data import *
import copy
import os

class Error(Exception):
    def __init__(self, message):
        self.message = message

class GridError(Exception):
    def __init__(self, row, col, message):
        self.message = f"Error in row {row}, column {col}: {message}"

def _transfer_position(position):
    if position == 'top':
        return 'north'
    if position == 'bottom':
        return 'south'
    if position == 'left':
        return 'west'
    if position == 'right':
        return 'east'

    if position in ['north', 'east', 'south', 'west']:
        return position

    raise Error('Incorrect position. Try: "top"/"left"/... or "north"/"west"/...')

def _is_north_or_south(pos):
    if pos in ['north', 'south', 'top', 'bottom']:
        return True
    return False

class LayoutView:
    '''
    This class is used to make changes in a 'Grid' layout to be more user friendly.
    '''

    def __init__(self, grid):
        self.layout = grid.data['layout']

    def set_padding(self, top=None, left=None, bottom=None, right=None, column=None, row=None):
        '''
        unit: mm (float) for top/left/bottom/right, same for column/row
        '''
        if top is not None:
            self.layout['padding']['north'] = top
        if left is not None:
            self.layout['padding']['west'] = left
        if bottom is not None:
            self.layout['padding']['south'] = bottom
        if right is not None:
            self.layout['padding']['east'] = right
        if column is not None:
            self.layout['column_space'] = column
        if row is not None:
            self.layout['row_space'] = row
        return self

    def _set_text_properties(self, field, position, field_size_mm, offset_mm=None, fontsize=None,
                             txt_rotation=None, txt_color=None, line_space=None):
        if _is_north_or_south(position):
            field["height"] = field_size_mm
        else:
            field["width"] = field_size_mm

        if offset_mm is not None:
            field["offset"] = offset_mm
        if txt_rotation is not None:
            field["rotation"] = txt_rotation
        if fontsize is not None:
            field["fontsize"] = fontsize
        if txt_color is not None:
            field["text_color"] = txt_color
        if line_space is not None:
            field["line_space"] = line_space
        return self

    def set_caption(self, height_mm, offset_mm=None, fontsize=None, txt_rotation=None, txt_color=None, line_space=None):
        '''
        Writes text below/south (each) image of that grid.
        '''
        field = self.layout["element_config"]["captions"]["south"]
        self._set_text_properties(field, 'south', height_mm, offset_mm, fontsize, txt_rotation, txt_color, line_space)
        return self

    def set_title(self, position, field_size_mm, offset_mm=None, fontsize=None, txt_rotation=None,
                  txt_color=None, line_space=None, bg_color=None):
        '''
        Writes a grid/subfigure title.
        '''

        field = self.layout['titles'][_transfer_position(position)]
        self._set_text_properties(field, position, field_size_mm, offset_mm, fontsize, txt_rotation,
            txt_color, line_space)
        if bg_color is not None:
            field["background_color"] = bg_color
        return self

    def set_row_titles(self, position, field_size_mm, offset_mm=None, fontsize=None, txt_rotation=None,
                       txt_color=None, line_space=None, bg_color=None):
        field = self.layout['row_titles'][_transfer_position(position)]
        self._set_text_properties(field, position, field_size_mm, offset_mm, fontsize, txt_rotation,
            txt_color, line_space)
        if bg_color is not None:
            field["background_colors"] = bg_color
        return self

    def set_col_titles(self, position, field_size_mm, offset_mm=None, fontsize=None, txt_rotation=None,
                       txt_color=None, line_space=None, bg_color=None):
        field = self.layout['column_titles'][_transfer_position(position)]
        self._set_text_properties(field, position, field_size_mm, offset_mm, fontsize, txt_rotation,
            txt_color, line_space)
        if bg_color is not None:
            field["background_colors"] = bg_color
        return self

    def _set_field_size_if_not_set(self, component, pos, field_size_mm=5):
        '''
        Makes sure, that the corresponding field_size of new added content will be set (not zero) and, therefore,
        visible for the user.
        '''
        if _is_north_or_south(pos):
            field_size = component[pos]['height']
        else:
            field_size = component[pos]['width']
        if field_size is None or field_size == 0:
            self._set_text_properties(component[pos], pos, field_size_mm)

class ElementView:
    '''
    A 'Grid' contains one or multiple elements depending on num_row and num_col.
    This class will help make changes in the settings for each element.
    You should however 'set_images' for each element, else unknown behaviour.
    '''
    def __init__(self, grid, row, col):
        self.elem = grid.data["elements"][row][col]
        self.row = row
        self.col = col
        self.layout = grid.get_layout()

    @property
    def image(self) -> ElementData:
        return self.elem['image']

    @image.setter
    def image(self, image: ElementData):
        self.set_image(image)

    def set_image(self, image: ElementData):
        if not isinstance(image, ElementData):
            try:
                image = PNG(image)
            except:
                raise GridError(self.row, self.col, 'set_image needs an image of type figuregen.Image (e.g. figuregen.PNG)'
                    'or of type figuregen.Plot (e.g. figuregen.MatplotLinePlot)')
            print("Deprecation warning: interpreted image raw data as figuregen.PNG")
        self.elem['image'] = image
        return self

    def set_frame(self, linewidth, color=[0,0,0]):
        '''
        linewidth (float): unit is in pt
        color: [r,g,b] each channel with int range (0-255)
        '''
        self.elem['frame'] = { "line_width": linewidth, "color": color }
        return self

    def draw_lines(self, start_positions, end_positions, linewidth_pt=0.5, color=[0,0,0]):
        '''
        start_positions/end_positions (list of tuples): defines the position of the line to draw
            on top of the image. Needs a tuple (x: row, y: column) in pixel.
        '''
        # Validate arguments
        if linewidth_pt <= 0.0:
            raise GridError(self.row, self.col, f'invalid linewidth: {linewidth_pt}. Please choose a '
                    'positive linewidth_pt > 0.')
        if not isinstance(start_positions, list) or start_positions == []:
            raise GridError(self.row, self.col, 'Invalid argument "start_positions" needs to be a '
                    f'list that is not empty. Given: {start_positions}.')
        if not isinstance(end_positions, list) or end_positions == []:
            raise GridError(self.row, self.col, 'Invalid argument "end_positions" needs to be a '
                    f'list that is not empty. Given: {end_positions}.')
        if len(start_positions) != len(end_positions):
            raise GridError(self.row, self.col, 'You have more start positions than end positions (or reverse).')
        if len(start_positions[0]) != 2 or len(end_positions[0]) != 2:
            raise GridError(self.row, self.col, 'Invalid argument "start_positions"/"end_positions" should'
                'be a list of tuples. Each tuple represents the x and y coordination in pixels.')

        try:
            self.elem["lines"].append({"from": start_positions[0], "to": end_positions[0], "color": color, "linewidth": linewidth_pt})
        except:
            self.elem["lines"] = []
            self.elem["lines"].append({"from": start_positions[0], "to": end_positions[0], "color": color, "linewidth": linewidth_pt})

        for i in range(1, len(start_positions)):
            self.elem["lines"].append({"from": start_positions[i], "to": end_positions[i], "color": color, "linewidth": linewidth_pt})

    def set_marker_properties(self, linewidth=1.5, is_dashed=False):
        print('Warning, function does not change marker properties anymore: set_marker_properties got replaced by set_marker.')
        return self

    def set_marker(self, pos, size, color=[255,255,255], linewidth_pt=1.0, is_dashed=False):
        '''
            Draws a rectangle on top of an image.

            args:
                pos (tuple): starting position (left, top) in pixel
                size (tuple): size of the rectangle (width, height) in pixel
        '''
        if linewidth_pt <= 0.0:
            raise Error('set_marker: invalid linewidth "'+str(linewidth_pt)+'". Please choose a positive linewidth_pt > 0.')

        try:
            test = self.elem["crop_marker"][0]
        except:
            self.elem["crop_marker"] = []

        self.elem["crop_marker"].append({"pos": pos, "size": size, "color": color, "linewidth": linewidth_pt, "dashed": is_dashed})
        return self

    def set_caption(self, txt_content):
        '''
        A (south) caption is placed below an image.
        In case the corresponding field height is not set yet, we set a 'default' value. This makes sure, that
        the content (provided by the user) will be shown. The user can set/overwrite the layout for captions anytime.
        '''
        self.elem["captions"] = {}
        self.elem["captions"]["south"] = str(txt_content)

        # check if caption layout is already set, if not, set a field_size,
        # so that the user is not confused, why content isn't shown
        self.layout._set_field_size_if_not_set(self.layout.layout['element_config']['captions'], pos='south', field_size_mm=6.)
        return self

    def set_label(self, txt_content, pos, width_mm=10., height_mm=3.0, offset_mm=[1.0, 1.0],
                  fontsize=6, bg_color=None, txt_color=[0,0,0], txt_padding_mm=1.0):
        '''
            Write text on top of an image.

            args:
                pos (str): e.g. 'bottom_right', 'top_left', 'top_center' (= 'top'), ...
                offset_mm (tuple): defines where the label is placed exactly.
                        We recommend to set bg_color, if you want to experiment with offsets.
                fontsize (float): unit point
                bg/txt_color (list): rgb integers ranging from 0 to 255.
        '''

        if not(pos in ['bottom', 'top', 'bottom_left', 'bottom_right', 'bottom_center', 'top_left', 'top_right', 'top_center']):
            raise Error("Label position '"+ pos +"' is invalid. Valid positions are: 'bottom_left',"
                "'bottom_right', 'bottom_center' (= 'bottom'), 'top_left', 'top_right', or 'top_center' (= 'top').")
        if pos == 'bottom' or pos == 'top':
            pos += '_center'

        try:
            self.elem["label"][pos] = {}
        except:
            self.elem["label"] = {}

        if 'center' in pos:
            try:
                offset_mm = offset_mm[0]
            except:
                pass

        self.elem["label"][pos] = {
            "text": str(txt_content),
            "fontsize": fontsize,
            "line_space": 1.2,
            "text_color": txt_color,
            "background_color": bg_color,
            "width_mm": width_mm,
            "height_mm": height_mm,
            "offset_mm": offset_mm,
            "padding_mm": txt_padding_mm
        }


class Grid:
    def __init__(self, num_rows, num_cols):
        ''' Create an empty grid
        '''
        self.data = {
            "elements": [[{} for i in range(num_cols)] for i in range(num_rows)],
            "row_titles": {},
            "column_titles": {},
            "titles": {},
            "layout": copy.deepcopy(_default_layout["grid"])
        }
        self.rows = num_rows
        self.cols = num_cols

    def __getitem__(self, rowcol) -> ElementView:
        return self.get_element(*rowcol)

    def get_element(self, row, col) -> ElementView:
        return ElementView(self, row, col)

    @property
    def aspect_ratio(self):
        """ Aspect ratio (height / width) of all images in the grid.
        Currently assumes that the user set them correctly such that they are all equal to
        the top left image.
        """
        return self[0, 0].image.aspect_ratio

    @property
    def layout(self):
        return self.get_layout()

    def get_layout(self):
        return LayoutView(self)

    def copy_layout(self, other: 'Grid') -> None:
        ''' Copies the layout of another grid. Useful to quickly align paddings and font settings.

        Args:
            other: the Grid object to copy the layout from
        '''
        assert isinstance(other, Grid)
        self.data["layout"] = copy.deepcopy(other.data["layout"])

    def set_title(self, position, txt_content) -> 'Grid':
        '''
        If the user has not specified the title size in the layout yet, it will be set to a default value of 6mm.

        Args:
            position: one of 'north'/'west'/... or 'top'/'right'/...

        Returns:
            This object (for chaining purposes)
        '''
        pos = _transfer_position(position)
        self.data["titles"][pos] = str(txt_content)

        # set a field_size (if not already done), so that the user is not confused, why content isn't shown
        self.get_layout()._set_field_size_if_not_set(self.layout.layout['titles'], pos=pos, field_size_mm=6.)
        return self

    def set_row_titles(self, position: str, txt_list: list):
        '''
        Args:
            position: string (valid: 'west'/'east' or 'right'/'left')
            txt_list: string list with one title for each row
        '''
        pos = _transfer_position(position)
        if pos in ['north', 'south']:
            raise Error("Invalid position for row_title. Try: 'west'/'east' or 'right'/'left'")

        assert isinstance(txt_list, list), "Please provide a list of strings, not a simple string."
        assert len(txt_list) >= self.rows, "Please provide a title for every row."

        try:
            self.data['row_titles'][pos]['content'] = txt_list
        except:
            self.data['row_titles'][pos] = {}
            self.data['row_titles'][pos]['content'] = txt_list

        # set a field_size (if not already done), so that the user is not confused, why content isn't shown
        self.get_layout()._set_field_size_if_not_set(self.layout.layout['row_titles'], pos=pos, field_size_mm=3.)
        return self

    def set_col_titles(self, position, txt_list):
        '''
        position: string (valid: 'north'/'south' or 'top'/'bottom')
        txt_list: string list of num_cols
        '''
        pos = _transfer_position(position)
        if pos in ['west', 'east']:
            raise Error('Invalid position for column_title. Try: "north"/"south" or "top"/"bottom"')
        if not isinstance(txt_list, list):
            raise Error ("'set_col_titles': Please give a list of strings, not a simple string. The length of the list should cover the number of columns.")
        if len(txt_list) < self.cols:
            raise Error ("'set_col_titles': length of provided list is less than number of columns.")

        try:
            self.data['column_titles'][pos]['content'] = txt_list
        except:
            self.data['column_titles'][pos] = {}
            self.data['column_titles'][pos]['content'] = txt_list

         # set a field_size (if not already done), so that the user is not confused, why content isn't shown
        self.get_layout()._set_field_size_if_not_set(self.layout.layout['column_titles'], pos=pos, field_size_mm=3.)
        return self

from .backend import Backend
from .tikz import TikzBackend
from .pdflatex import PdfBackend
from .html import HtmlBackend
from .powerpoint import PptxBackend

def _backend_from_filename(filename: str) -> Backend:
    """ Guesses the correct backend based on the filename """
    extension = os.path.splitext(filename)[1].lower()
    if extension == ".pptx":
        return PptxBackend()
    elif extension == ".html":
        return HtmlBackend()
    elif extension == ".pdf":
        return PdfBackend()
    elif extension == ".tikz":
        return TikzBackend()
    else:
        raise ValueError(f"Could not derive backend from extension '{filename}'. Please specify.")

def figure(grids: List[List[Grid]], width_cm: float, filename: str, backend: Backend = None):
    """
    Grid rows: Creates a figure by putting grids next to each other, from left to right.
    Grid columns: stacks rows vertically.
    Aligns the height of the given grids such that they fit the given total width.

    Args:
        grids: a list of lists of Grids (figuregen.Grid), which stacks horizontal figures vertically
        width_cm: total width of the figure in centimeters
        backend: a Backend object that will be used to create the figure, or None to use a default
    """
    if backend is None:
        backend = _backend_from_filename(filename)
    backend.generate(grids, width_cm * 10, filename)

def horizontal_figure(grids, width_cm: float, filename, backend: Backend = None):
    """
    Creates a figure by putting grids next to each other, from left to right.
    Aligns the height of the given grids such that they fit the given total width.

    Args:
        grids: a list of Grids (figuregen.Grid)
        width_cm: total width of the figure in centimeters
        backend: a Backend object that will be used to create the figure, or None to use a default
    """
    figure([grids], width_cm, filename, backend)